/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifdef HAVE_CONFIG_H
    #include "autoconfig.h"
#endif

#include <boost/pointer_cast.hpp>
#include <errno.h>
#include <string.h>
#include <exception>
#include <iomanip>

#include "log.h"
#include "util.h"
#include "exceptions.h"
#include "config_manager.h"
#include "hwdesc.h"
#include "hwstats.h"
#include "storage.h"
#include "database.h"
#include "ds-assistant.h"
#include "database.h"
#include "util.h"
#include "traffic-decode.h"

#define INVALID_TASK_ID 0

namespace dSA
{

dS_Assistant* dS_Assistant::m_instance = NULL;

dS_Assistant* dS_Assistant::createInstance()
{
    if (m_instance == NULL)
    {
        m_instance = new dS_Assistant();
        if (m_instance == NULL)
        {
            throw _Exception("could not create dS Assistant class");
        }
    }

    return m_instance;
}

dS_Assistant* dS_Assistant::getInstance()
{
    if (m_instance == NULL)
    {
        throw _Exception("dS Assistant not initialized");
    }

    return m_instance;
}
dS_Assistant::dS_Assistant() : m_shutdown_flag(false),
                               m_running(false),
                               m_task_thread(0),
                               m_long_task_thread(0)
{
    int ret;

    ConfigManager *cfg = ConfigManager::getInstance();
    std::string logdir(cfg->getOption(CFG_LOG_DIR));
    m_logfile_name = logdir +  + "/dsa.log";
    m_logfile = fopen(m_logfile_name.c_str(), "a");
    if (NULL == m_logfile)
    {
        throw _Exception("failed to open logfile \'" + m_logfile_name + "\': " +
                         std::string(strerror(errno)));
    }

    Output2DSID::SetLogdirectory(logdir);
    DSIDLog::ReportingLevel() =
        DSIDLog::FromString((cfg->getOption(CFG_LOG_LEVEL)).c_str());

    Output2FILE::SetStream(m_logfile);
    FILELog::ReportingLevel() =
        FILELog::FromString((cfg->getOption(CFG_LOG_LEVEL)).c_str());
    FILE_LOG(logINFO) << "dS Assistant startup";

    SetNullId(m_own_dsid);

    ret = pthread_mutex_init(&m_task_list_mutex, NULL);
    if (ret != 0)
    {
        throw _Exception("failed to initialize task list mutex: " +
                         std::string(strerror(errno)));
    }

    ret = pthread_mutex_init(&m_long_task_list_mutex, NULL);
    if (ret != 0)
    {
        pthread_mutex_destroy(&m_task_list_mutex);
        throw _Exception("failed to initialize long task list mutex: " +
                         std::string(strerror(errno)));
    }

    ret = pthread_cond_init(&m_wakeup_task_cond, NULL);
    if (ret != 0)
    {
        pthread_mutex_destroy(&m_task_list_mutex);
        pthread_mutex_destroy(&m_long_task_list_mutex);
        throw _Exception("failed to initialize task thread wakeup condition: " +
                         std::string(strerror(errno)));
    }

    ret = pthread_cond_init(&m_wakeup_long_task_cond, NULL);
    if (ret != 0)
    {
        pthread_mutex_destroy(&m_task_list_mutex);
        pthread_mutex_destroy(&m_long_task_list_mutex);
        pthread_cond_destroy(&m_wakeup_task_cond);
        throw _Exception("failed to initialize task thread wakeup condition: " +
                         std::string(strerror(errno)));
    }
}

void dS_Assistant::reopenLog()
{
    Output2FILE::SetStream(NULL);
    if (NULL != m_logfile)
    {
        fclose(m_logfile);
    }
    m_logfile = fopen(m_logfile_name.c_str(), "a");
    if (NULL == m_logfile)
    {
        throw _Exception("failed to re-open logfile \'" + m_logfile_name + "\': " +
                         std::string(strerror(errno)));
    }
    Output2FILE::SetStream(m_logfile);
}

void dS_Assistant::loadDatabaseDefaults()
{
    // query default database settings
    std::string apartment_id;
    try {
        apartment_id = Database::getInstance()->getConfig("ApartmentId");
    }
    catch (Exception& ex ) {
        std::string systemid = util::mac2str();
        char c = systemid.length() ? systemid[systemid.length() - 1] : '0';
        uint8_t apartmentid_db = (strtol(&c, NULL, 16) & 7) * 8;

        FILE_LOG(logINFO) << "Set apartment id from system mac address "
                << systemid << " to " << util::str(apartmentid_db);
        try
        {
            Database::getInstance()->setConfig("ApartmentId",
                    util::str(apartmentid_db));
        }
        catch (Exception& ex)
        {
            FILE_LOG(logERROR) << "Failed to store apartment id in database: "
                    << ex.getMessage();
        }
    }
}

void dS_Assistant::connect()
{
    boost::mutex::scoped_lock lock(m_dsa_mutex);

    if (m_dsm_api_wrapper != NULL)
    {
        throw _Exception("connection already exists");
    }

    ConfigManager *cfg = ConfigManager::getInstance();
    m_dsm_api_wrapper = boost::shared_ptr<dSMApiWrapper>(
                    new dSMApiWrapper(cfg->getOption(CFG_DS485_DAEMON_URL)));

    try
    {
        m_dsm_api_wrapper->getOwnDSID(&m_own_dsid);
    }
    catch (Exception ex)
    {
        m_dsm_api_wrapper.reset();
        throw ex;
    }

    m_bus_state = m_dsm_api_wrapper->getBusState();

    m_dsm_api_wrapper->addListener(this);
}

bus_state_t dS_Assistant::getBusState() const
{
    boost::mutex::scoped_lock lock(m_dsa_mutex);
    return m_bus_state;
}

boost::shared_ptr<dSMApiWrapper> dS_Assistant::dS485()
{
    if (m_dsm_api_wrapper == NULL)
    {
        throw _Exception("no active connection to dS485 daemon available");
    }
    return m_dsm_api_wrapper;
}

void dS_Assistant::addTask(boost::shared_ptr<dSTask> task)
{
    if (task == NULL)
    {
        throw _Exception("Will not add invalid NULL task");
    }
    lockTaskList();
    m_task_list.push_front(task);
    pthread_cond_signal(&m_wakeup_task_cond);
    unlockTaskList();
}

long long int dS_Assistant::addLongTask(boost::shared_ptr<dSLongTask> task)
{
    lockLongTaskList();
#ifndef ADVANCED_TASK_MANAGEMENT
    if (m_current_task_info != NULL)
    {
        m_current_task_info->invalidate();
    }
#endif
    long long int task_id = Database::getInstance()->saveTask(task);
    task->setTaskID(task_id);

    m_long_task_list.push_front(task);
    pthread_cond_signal(&m_wakeup_long_task_cond);
    unlockLongTaskList();
    return task_id;
}

///****************************************************************************

void dS_Assistant::busStateCallback(bus_state_t state)
{
    m_bus_state = state;
    FILE_LOG(logINFO) << "dS485 Bus State: " << state;
}


void dS_Assistant::busChangeCallback(dsid_t *id, int flag)
{
    try
    {
        if (!DsmApiIsdSM(*id))
        {
            return;
        }

        FILE_LOG(logDEBUG1) << "dS485 Bus Change: " << util::str(*id) <<
                               " - " << flag;

        boost::shared_ptr<const dSM> dsmo = Storage::getInstance()->getDSM(*id);
        boost::shared_ptr<dSM> dsm(dsmo == NULL ? new dSM(*id) : dsmo->clone());

        if (flag)
        {
            DSID_LOG(logINFO, util::str(*id)) << "left";
            dsm->setPresent(0);
        }
        else
        {
            DSID_LOG(logINFO, util::str(*id)) << "joined";
            dsm->setPresent(1);
            boost::shared_ptr<GetDSMInfoTask> task(
                    new GetDSMInfoTask(dsm->getDSID()));
            addTask(boost::dynamic_pointer_cast<dSTask>(task));
        }
        Storage::getInstance()->addDSM(dsm);
    }
    catch (Exception ex)
    {
        FILE_LOG(logERROR) << "Exception in busChangeCallback: " <<
                              ex.getMessage();
        //FILE_LOG(logERROR) << ex.getStackTrace();
    }
    catch (std::exception e)
    {
        FILE_LOG(logERROR) << "Exception in busChangeCallback: " << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR) << "Exception in busChangeCallback";
    }
}

///****************************************************************************

void dS_Assistant::deviceAddedCallback(uint8_t error_code,
                                       dsid_t src_dsm_dsid,
                                       dsid_t dest_dsm_dsid,
                                       uint16_t device_id,
                                       uint16_t zone_id,
                                       uint16_t vendor_id,
                                       uint32_t serial_number)
{
    dsid_t dsid;
    if (DsmApiExpandDeviceDSID(vendor_id, serial_number, &dsid) != 0)
    {
        FILE_LOG(logDEBUG2) << "Ignoring unidentified device on dSM " <<
                               util::str(src_dsm_dsid);
        return;
    }

    boost::shared_ptr<const Device> dsdo =
                        Storage::getInstance()->getDevice(dsid, src_dsm_dsid);
    if (dsdo)
    {
        boost::shared_ptr<Device> dsd(dsdo->clone());
        dsd->setState(1);
        Storage::getInstance()->addDevice(dsd);
    }

    boost::shared_ptr<GetAllDevicesTask> task(new GetAllDevicesTask(src_dsm_dsid));
    addTask(boost::dynamic_pointer_cast<dSTask>(task));

    DSID_LOG(logINFO, util::str(dsid)) << "active";
    FILE_LOG(logINFO) << "Added device with dSID " <<
            util::str(dsid) << " and address " << util::str(device_id) <<
            " on dSM " << util::str(src_dsm_dsid);
}

void dS_Assistant::deviceRemovedCallback(uint8_t error_code,
                                         dsid_t src_dsm_dsid,
                                         dsid_t dest_dsm_dsid,
                                         uint16_t device_id,
                                         uint16_t zone_id,
                                         uint16_t vendor_id,
                                         uint32_t serial_number)
{
    dsid_t dsid;
    if (DsmApiExpandDeviceDSID(vendor_id, serial_number, &dsid) != 0)
    {
        FILE_LOG(logDEBUG2) << "Ignoring unidentified device on dSM " <<
                               util::str(src_dsm_dsid);
        return;
    }

    boost::shared_ptr<const Device> dsdo =
                        Storage::getInstance()->getDevice(dsid, src_dsm_dsid);
    if (dsdo)
    {
        boost::shared_ptr<Device> dsd(dsdo->clone());
        dsd->setState(0);
        Storage::getInstance()->addDevice(dsd);
    }
    DSID_LOG(logINFO, util::str(dsid)) << "inactive";
    FILE_LOG(logINFO) << "Removed device with dSID " <<
            util::str(dsid) << " and address " << util::str(device_id) <<
            " on dSM " << util::str(src_dsm_dsid);
}

void dS_Assistant::deviceStatisticsCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t offset, uint16_t value)
{
    boost::shared_ptr<const Device> dsdo =
                Storage::getInstance()->getDevice(device_id, src_dsm_dsid);
    if (dsdo)
    {
        DSID_PREFIX_LOG(logDEBUG, dsdo->getDSID(), "stat_") <<
            "statistic offset:" << util::hex(offset,2) <<
            ", value:" << util::hex(value,4);
    }
}

void dS_Assistant::deviceActionCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint16_t zone_id, uint8_t group_id, uint8_t button_nr,
        uint8_t click_type)
{
}

void dS_Assistant::deviceLocalActionCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint16_t zone_id, uint8_t state)
{
}

void dS_Assistant::deviceConfigCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t config_class, uint8_t config_index, uint8_t value)
{
    boost::shared_ptr<const Device> dsdo =
                    Storage::getInstance()->getDevice(device_id, src_dsm_dsid);
    if (dsdo)
    {
        DSID_LOG(logINFO, util::str(dsdo->getDSID())) <<
            "device config class:" << util::hex(config_class) <<
            ", index:" << util::hex(config_index, 2) <<
            ", value:" << util::hex(value, 4);
    }
}

void dS_Assistant::deviceTestTransmissionQualityCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t downstream_quality, uint16_t upstream_quality)
{
    boost::shared_ptr<const Device> dsdo =
        Storage::getInstance()->getDevice(device_id, src_dsm_dsid);
    if (dsdo)
    {
        DSID_LOG(logINFO, util::str(dsdo->getDSID())) <<
            "test quality down:" << util::hex(downstream_quality,2) <<
            ", up:" << util::hex(upstream_quality,2);
    }
}

///****************************************************************************

void dS_Assistant::commUpstreamShortCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
        uint8_t key_type, uint8_t click_type,
        uint8_t quality, uint8_t flags)
{
    DSID_LOG(logINFO, util::str(src_dsm_dsid)) <<
        "[Q:" << std::setprecision (2) <<
        (quality * 100.0/62 + .5) << ", P:" << ((flags & 1) > 0 ? 'N' : 'P') <<
        "] " <<
        dsutil::rkDecodeShort(src_dsm_dsid, device_id, is_resend, is_sensor,
                              key_type, click_type);
}

void dS_Assistant::commUpstreamLongCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
        uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier,
        uint32_t data, uint8_t quality, uint8_t flags)
{
    DSID_LOG(logINFO, util::str(src_dsm_dsid)) <<
        "[Q:" << std::setprecision (2) <<
        (quality * 100.0/62 + .5) << ", P:" <<
        ((flags & 1) > 0 ? 'N' : 'P') << "] " <<
        dsutil::rkDecodeLong(src_dsm_dsid, device_id, mainqualifier,
                             subqualifier, data);
}

void dS_Assistant::commDownstreamShortCallback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command)
{
    DSID_LOG(logINFO, util::str(src_dsm_dsid)) <<
            dsutil::hkDecodeShort(command);
}

void dS_Assistant::commDownstreamLong0Callback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command)
{
    DSID_LOG(logINFO, util::str(src_dsm_dsid)) <<
            dsutil::hkDecodeLong0(command);
}

void dS_Assistant::commDownstreamLong8Callback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command,
        uint8_t data)
{
    DSID_LOG(logINFO, util::str(src_dsm_dsid)) <<
            dsutil::hkDecodeLong8(command, data);
}

void dS_Assistant::commDownstreamLong16Callback(uint8_t error_code,
        dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command,
        uint16_t data)
{
    DSID_LOG(logINFO, util::str(src_dsm_dsid)) <<
            dsutil::hkDecodeLong16(command, data);
}

///****************************************************************************

void dS_Assistant::lockTaskList()
{
    int ret = pthread_mutex_lock(&m_task_list_mutex);
    if (ret != 0)
    {
        throw _Exception("Failed to lock task list mutex: " +
                         std::string(strerror(errno)));
    }
}

void dS_Assistant::unlockTaskList()
{
    int ret = pthread_mutex_unlock(&m_task_list_mutex);
    if (ret != 0)
    {
        throw _Exception("Failed to unlock task list mutex: " +
                         std::string(strerror(errno)));
    }
}

void dS_Assistant::lockLongTaskList() const
{
    int ret = pthread_mutex_lock(&m_long_task_list_mutex);
    if (ret != 0)
    {
        throw _Exception("Failed to lock long task list mutex: " +
                         std::string(strerror(errno)));
    }
}

void dS_Assistant::unlockLongTaskList() const
{
    int ret = pthread_mutex_unlock(&m_long_task_list_mutex);
    if (ret != 0)
    {
        throw _Exception("Failed to unlock long task list mutex: " +
                         std::string(strerror(errno)));
    }
}

void dS_Assistant::run()
{
    int ret;

    if (m_running)
    {
        throw _Exception("Task threads are already running");
    }

    loadDatabaseDefaults();
    connect();

    m_running = true;

    try
    {
        clearPendingTasks();
    }
    catch (Exception ex)
    {
        m_running = false;
        throw ex;
    }

    ret = pthread_create(&m_task_thread, NULL, dS_Assistant::staticThreadProc,
                         this);
    if (ret != 0)
    {
        disconnect();
        m_running = false;
        throw _Exception("Could not start task thread");
    }

    ret = pthread_create(&m_long_task_thread, NULL,
                         dS_Assistant::staticLongTaskThreadProc, this);
    if (ret != 0)
    {
        pthread_join(m_task_thread, NULL);
        disconnect();
        m_running = false;
        throw _Exception("Could not start long task thread");
    }

    Database::getInstance()->updateTaskStatus(Running, Cancelled);
#ifdef ADVANCED_TASK_MANAGEMENT
    scan(false); // synchronous scan

    boost::shared_ptr<std::list<boost::shared_ptr<dSLongTask> > > tasks =
        Database::getInstance()->getTasksByStatus(Queued);

    if (tasks->empty())
    {
        return;
    }

    lockLongTaskList();
    while (!tasks->empty())
    {
        boost::shared_ptr<dSLongTask> task = tasks->front();
        tasks->pop_front();
        m_long_task_list.push_front(task);

    }
    pthread_cond_signal(&m_wakeup_long_task_cond);
    unlockLongTaskList();
#else
    Database::getInstance()->removeTasksByStatus(Queued);
    scan(true); // asynchronous scan
#endif
}

void dS_Assistant::taskThread()
{
    boost::shared_ptr<dSTask> task;

    lockTaskList();
    while (!m_shutdown_flag)
    {
        if (m_task_list.empty())
        {
            pthread_cond_wait(&m_wakeup_task_cond, &m_task_list_mutex);
            continue;
        }

        task = m_task_list.back();
        m_task_list.pop_back();

        unlockTaskList();

        try
        {
            if ((task != NULL) && (task->isValid()))
            {
                task->run(this);
            }
        }
        catch (Exception ex)
        {
            FILE_LOG(logERROR) << "Task thread caught exception: " <<
                                  ex.getMessage();
            ex.printStackTrace();
        }
        catch (std::exception e)
        {
            FILE_LOG(logERROR) << "Task thread caught std exception: " <<
                                  e.what();
        }
        catch (...)
        {
            FILE_LOG(logERROR) << "Task thread caught unknown exception";
        }

        task.reset();

        if (!m_shutdown_flag)
        {
            lockTaskList();
        }
    }
    unlockTaskList();
}

void *dS_Assistant::staticThreadProc(void *arg)
{
    dS_Assistant *instance = (dS_Assistant *)arg;
    instance->taskThread();
    pthread_exit(NULL);
    return NULL;
}

void dS_Assistant::longTaskThread()
{
    boost::shared_ptr<dSLongTask> task;

    lockLongTaskList();
    while (!m_shutdown_flag)
    {
        if (m_long_task_list.empty())
        {
            pthread_cond_wait(&m_wakeup_long_task_cond,
                              &m_long_task_list_mutex);
            continue;
        }

        task = m_long_task_list.back();
        m_long_task_list.pop_back();

        m_current_task_info = boost::dynamic_pointer_cast<dSLongTaskInfo>(task);
        unlockLongTaskList();

        try
        {
            if (task->isValid())
            {
                Database::getInstance()->updateTaskStatus(task->getTaskID(),
                                                          Running);
                task->run(this);
                if (task->isValid())
                {
                    Database::getInstance()->updateTaskStatus(task->getTaskID(),
                                                              Finished);
                }
                else
                {
                    Database::getInstance()->updateTaskStatus(task->getTaskID(),
                                                              Cancelled);
                }
            }
        }
        catch (Exception ex)
        {
            FILE_LOG(logERROR) << "Long task thread caught exception: " <<
                ex.getMessage();
            ex.printStackTrace();
        }
        catch (std::exception e)
        {
            FILE_LOG(logERROR) << "Long task thread caught exception: " <<
                e.what();
        }
        catch (...)
        {
            FILE_LOG(logERROR) << "Long task thread caught unknown exception";
        }

        task.reset();

        if (!m_shutdown_flag)
        {
            lockLongTaskList();
            m_current_task_info.reset();
        }
    }
    unlockLongTaskList();
}

void *dS_Assistant::staticLongTaskThreadProc(void *arg)
{
    dS_Assistant *instance = (dS_Assistant *)arg;
    instance->longTaskThread();
    pthread_exit(NULL);
    return NULL;
}

void dS_Assistant::scan(bool async)
{
    boost::shared_ptr<GetBusMembersTask> task(new GetBusMembersTask(async));
    if (async)
    {
        addTask(boost::dynamic_pointer_cast<dSTask>(task));
    }
    else
    {
        task->run(this);
    }
}

void dS_Assistant::clearPendingTasks()
{
    lockTaskList();
    while (!m_task_list.empty())
    {
        boost::shared_ptr<dSTask> task = m_task_list.front();
        task.reset();
        m_task_list.pop_front();
    }
    unlockTaskList();

    lockLongTaskList();
    while (!m_long_task_list.empty())
    {
        boost::shared_ptr<dSLongTask> task = m_long_task_list.front();
        task.reset();
        m_long_task_list.pop_front();
    }
    unlockLongTaskList();
}

boost::shared_ptr<const dSLongTaskInfo> dS_Assistant::getCurrentTaskInfo() const
{
    lockLongTaskList();
    boost::shared_ptr<const dSLongTaskInfo> ret = m_current_task_info;
    unlockLongTaskList();
    return ret;
}

boost::shared_ptr<std::list<boost::shared_ptr<const dSLongTaskInfo> > >
                                        dS_Assistant::getTaskList() const
{
    lockLongTaskList();
    boost::shared_ptr<std::list<boost::shared_ptr<const dSLongTaskInfo> > >
        taskList(new std::list<boost::shared_ptr<const dSLongTaskInfo> >);

    if (m_current_task_info != NULL)
    {
        taskList->push_back(m_current_task_info);
    }

    for (std::list<boost::shared_ptr<dSLongTask> >::const_iterator
            it = m_long_task_list.begin(); it != m_long_task_list.end(); it++)
    {
        boost::shared_ptr<const dSLongTaskInfo> task =
                            boost::dynamic_pointer_cast<dSLongTaskInfo>(*it);
        taskList->push_back(task);
    }

    unlockLongTaskList();

    return taskList;
}

void dS_Assistant::cancelTask(long long int task_id)
{
    bool found = false;

    lockLongTaskList();
    if ((m_current_task_info != NULL) &&
       (m_current_task_info->getTaskID() == task_id))
    {
        found = true;
        m_current_task_info->invalidate();
    }
    else
    {
        for (std::list<boost::shared_ptr<dSLongTask> >::iterator
            it = m_long_task_list.begin(); it != m_long_task_list.end(); it++)
        {
            boost::shared_ptr<dSLongTask> task = *it;
            if (task->getTaskID() == task_id)
            {
                found = true;
                task->invalidate();
                Database::getInstance()->updateTaskStatus(task_id, Cancelled);
                m_long_task_list.erase(it);
                break;
            }
        }
    }
    unlockLongTaskList();
    if (!found)
    {
        throw _Exception("task with id " + util::str(task_id) + " not found");
    }
}

void dS_Assistant::disconnect()
{
    boost::mutex::scoped_lock lock(m_dsa_mutex);
    if (m_dsm_api_wrapper)
    {
        try
        {
            clearPendingTasks();
        }
        catch (Exception ex)
        {
            FILE_LOG(logERROR) << "Failed to clear pending tasks: " <<
                ex.getMessage();
        }
        m_dsm_api_wrapper->removeListener(this);
        m_dsm_api_wrapper.reset();
    }
}

void dS_Assistant::shutdown()
{
    boost::mutex::scoped_lock lock(m_dsa_mutex);
    m_shutdown_flag = true;

    clearPendingTasks();
    lockTaskList();
    pthread_cond_signal(&m_wakeup_task_cond);
    unlockTaskList();
    if (m_task_thread)
    {
        pthread_join(m_task_thread, NULL);
        m_task_thread = 0;
    }

    lockLongTaskList();
    if (m_current_task_info != NULL)
    {
        m_current_task_info->invalidate();
    }
    pthread_cond_signal(&m_wakeup_long_task_cond);
    unlockLongTaskList();
    if (m_long_task_thread)
    {
        pthread_join(m_long_task_thread, NULL);
        m_long_task_thread = 0;
    }

    lock.unlock();
    disconnect();
}

const dsid_t dS_Assistant::getDSID() const
{
    return m_own_dsid;
}

long long int dS_Assistant::getInvalidTaskID() const
{
    return INVALID_TASK_ID;
}
dS_Assistant::~dS_Assistant()
{
    boost::mutex::scoped_lock lock(m_dsa_mutex);
    m_instance = NULL;
    lock.unlock();
    try
    {
        shutdown();
        disconnect();
        Database::getInstance()->updateTaskStatus(Running, Cancelled);
#ifndef ADVANCED_TASK_MANAGEMENT
        Database::getInstance()->removeTasksByStatus(Queued);
#endif
    }
    catch (...)
    {
        FILE_LOG(logERROR) << "Exception caught when shutting down, ignoring";
    }

    pthread_mutex_destroy(&m_task_list_mutex);
    pthread_mutex_destroy(&m_long_task_list_mutex);
    pthread_cond_destroy(&m_wakeup_task_cond);
    pthread_cond_destroy(&m_wakeup_long_task_cond);
}

dSTaskInfo::dSTaskInfo() : m_name("Generic task"), m_description(""),
                           m_valid(false),
                           m_cancellable(true) {}

void dSTaskInfo::setDescription(std::string description)
{
    m_description = description;
}

const std::string dSTaskInfo::getName() const
{
    return m_name;
}

const std::string dSTaskInfo::getDescription() const
{
    return m_description;
}

bool dSTaskInfo::isValid() const
{
    return m_valid;
}

void dSTaskInfo::invalidate()
{
    if (m_cancellable)
    {
        m_valid = false;
    }
}

dSTaskInfo::~dSTaskInfo() {}

bool dSTaskInfo::isCancellable() const
{
    return m_cancellable;
}

dSTask::dSTask() : dSTaskInfo() {}

dSLongTaskInfo::dSLongTaskInfo() : dSTaskInfo(), m_task_id(INVALID_TASK_ID),
                                    m_parent_task_id(INVALID_TASK_ID),
                                    m_progress(0),
                                    m_status_info("Queued"),
                                    m_task_status(Queued),
                                    m_started(0),
                                    m_finished(0),
                                    m_queued(0) {}

void dSLongTaskInfo::setTaskID(long long int task_id)
{
    m_task_id = task_id;
}

long long int dSLongTaskInfo::getTaskID() const
{
    return m_task_id;
}

void dSLongTaskInfo::setParentTaskID(long long int task_id)
{
    m_parent_task_id = task_id;
}

long long int dSLongTaskInfo::getParentTaskID() const
{
    return m_parent_task_id;
}

unsigned char dSLongTaskInfo::getProgress() const
{
    boost::mutex::scoped_lock lock(m_status_mutex);
    return m_progress;
}

std::string dSLongTaskInfo::getStatusInfo() const
{
    boost::mutex::scoped_lock lock(m_status_mutex);
    return m_status_info;
}

std::string dSLongTaskInfo::getParameters() const
{
    return m_parameters;
}

void dSLongTaskInfo::setTaskStatus(task_status status)
{
    boost::mutex::scoped_lock lock(m_status_mutex);
    m_task_status = status;
    switch (status)
    {
        case Running:
            m_status_info = "läuft gerade";
            break;
        case Queued:
            m_status_info = "gequeued";
            break;
        case Cancelled:
            m_status_info = "abgebrochen";
            break;
        case Finished:
            m_status_info = "beendet";
            break;
        case UnknownStatus:
        default:
            m_status_info = "unbekannt";
            break;
    }
}

task_status dSLongTaskInfo::getTaskStatus() const
{
    boost::mutex::scoped_lock lock(m_status_mutex);
    return m_task_status;
}

std::string dSLongTaskInfo::getTaskStatusText() const
{
    boost::mutex::scoped_lock lock(m_status_mutex);
    if (m_task_status == Running)
    {
        return "running";
    }
    else if (m_task_status == Cancelled)
    {
        return "cancelled";
    }
    else if (m_task_status == Finished)
    {
        return "finished";
    }
    else if (m_task_status == Queued)
    {
        return "queued";
    }
    else
    {
        return "unknown";
    }
}

time_t dSLongTaskInfo::getStartedTime() const
{
    return m_started;
}

void dSLongTaskInfo::setStartedTime(time_t t)
{
    m_started = t;
}

time_t dSLongTaskInfo::getFinishedTime() const
{
    return m_finished;
}

void dSLongTaskInfo::setFinishedTime(time_t t)
{
    m_finished = t;
}

time_t dSLongTaskInfo::getQueuedTime() const
{
    return m_queued;
}

void dSLongTaskInfo::setQueuedTime(time_t t)
{
    m_queued = t;
}

dSLongTask::dSLongTask() : dSLongTaskInfo() {}

GetBusMembersTask::GetBusMembersTask(bool async) : dSTask(), m_async(async)
{
    m_name = "Get bus members";
    m_description = "Scan the dS485 bus in order to retrieve information "
                    "about all available devices";
    m_valid = true;
}


void GetBusMembersTask::run(dS_Assistant *dsa)
{
    dsid_t device_list[DSM_API_MAX_DEVICES];

    if (!(dsa && m_valid))
    {
        return;
    }

    int devices = dsa->dS485()->getBusMembers(device_list, DSM_API_MAX_DEVICES);
    for (int i = 0; i < devices; ++i)
    {
        dsid_t own_dsid = dsa->getDSID();
        if (IsEqualId(device_list[i], own_dsid))
        {
            continue;
        }

        if (DsmApiIsdSM(device_list[i]))
        {
            boost::shared_ptr<dSM> dsm(new dSM(device_list[i]));
            Storage::getInstance()->addDSM(dsm);
            boost::shared_ptr<GetDSMInfoTask> task(
                                  new GetDSMInfoTask(dsm->getDSID(), m_async));
            if (m_async)
            {
                dsa->addTask(boost::dynamic_pointer_cast<dSTask>(task));
            }
            else
            {
                task->run(dsa);
            }
        }
        else
        {
            boost::shared_ptr<HWBase> h(new HWBase(device_list[i]));
            Storage::getInstance()->add(h);
        }
    }
}

GetDSMInfoTask::GetDSMInfoTask(dsid_t dsid, bool async) : dSTask(),
                                                          m_dsid(dsid),
                                                          m_async(async)
{
    m_name = "Request dSM Information";
    m_description = "Retrieve information for dSM " + util::str(dsid);
    m_valid = true;
}

void GetDSMInfoTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    uint32_t hwv;
    uint32_t armv;
    uint32_t dspv;
    uint16_t apiv;
    dsid_t dsid;
    uint8_t name[21];
    uint32_t cHash, cVersion;
    uint64_t configId;

    dsa->dS485()->dSMInfo(m_dsid, &hwv, &armv, &dspv, &apiv, &dsid, name);
    dsa->dS485()->getConfigurationId(m_dsid, &cHash, &cVersion);
    configId = (uint64_t) cHash << 32 | cVersion;

    DSID_LOG(logINFO, m_dsid) << "[Task DSMInfo] ARM:" << util::hex(armv) <<
            ", DSP:" << util::hex(dspv) <<
            ", HW:" << util::hex(hwv) <<
            ", CFG:" << configId;

    boost::shared_ptr<const dSM> dsmo = Storage::getInstance()->getDSM(m_dsid);
    boost::shared_ptr<dSM> dsm(dsmo == NULL ? new dSM(m_dsid) : dsmo->clone());
    dsm->setPresent(1);
    dsm->setHWVersion(hwv);
    dsm->setARMSwVersion(armv);
    dsm->setDSPSwVersion(dspv);
    dsm->setAPIVersion(apiv);
    dsm->setName(std::string((const char*)name));
    Storage::getInstance()->addDSM(dsm);

    int apartmentid_db = 255;
    uint8_t apartmentid_dsm = 255;

    try {
        dsa->dS485()->getApartmentId(m_dsid, &apartmentid_dsm);
    }
    catch (Exception ex)
    {
        FILE_LOG(logERROR) << "Failed to query apartment id from dsm "
            << util::str(m_dsid);
    }

    try
    {
        apartmentid_db = util::str2int(
                Database::getInstance()->getConfig("ApartmentId"));
    }
    catch (Exception ex)
    {
        FILE_LOG(logERROR) << "Failed to query apartment id from database: "
                << ex.getMessage();
    }

    if (((uint8_t) apartmentid_db < 63) && (apartmentid_dsm < 63) &&
        ((uint8_t) apartmentid_db != apartmentid_dsm))
    {
        boost::shared_ptr<SetDSMApartmentIdTask> task(
                new SetDSMApartmentIdTask(m_dsid, apartmentid_db, m_async));
        task->run(dsa);
        return;
    }

    boost::shared_ptr<StartupDSMTask> task(new StartupDSMTask(m_dsid, false));
    task->run(dsa);

    dsmo = Storage::getInstance()->getDSM(m_dsid);
    boost::shared_ptr<dSM> dsmconfig(dsmo->clone());

    dsmconfig->setApartmentId(apartmentid_dsm);
    if (configId != dsmconfig->getConfigurationId())
    {
        boost::shared_ptr<ScanDSMTask> task(new ScanDSMTask(m_dsid, m_async));
        dsmconfig->setConfigurationId(configId);
        if (m_async)
        {
            dsa->addTask(boost::dynamic_pointer_cast<dSTask>(task));
        }
        else
        {
            task->run(dsa);
        }
    }
    Storage::getInstance()->addDSM(dsmconfig);
}

ScanDSMTask::ScanDSMTask(dsid_t dsid, bool async) : dSTask(), m_dsid(dsid),
                                                    m_async(async)
{
    m_name = "Scanning dSM for devices";
    m_description = "Retrieve information about devices attached to the "
                    "dS-Meter";
    m_valid = true;
}

void ScanDSMTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    boost::shared_ptr<const dSM> dsmo = Storage::getInstance()->getDSM(m_dsid);
    if (dsmo == NULL)
    {
        throw _Exception("dSM " + util::str(m_dsid) + " not found");
    }
    boost::shared_ptr<dSM> dsm(dsmo->clone());

    uint16_t num;
    dsa->dS485()->getDevicesCountInZone(m_dsid, 0, &num);
    dsm->setDeviceCount(num);
    dsa->dS485()->getActiveDevicesCountInZone(m_dsid, 0, &num);
    dsm->setActiveDeviceCount(num);
    dsa->dS485()->getInactiveDevicesCountInZone(m_dsid, 0, &num);

    uint8_t count;
    dsa->dS485()->getZoneCount(m_dsid, &count);

    for (uint8_t i = 0; i < count; i++)
    {
        uint16_t zone_id;
        dsa->dS485()->getZoneInfo(m_dsid, i, &zone_id, NULL, NULL, NULL);
        dsm->setZoneId(i, zone_id);
    }

    // update storage
    Storage::getInstance()->addDSM(dsm);

    // schedule device scanner task
    boost::shared_ptr<GetAllDevicesTask> task(new GetAllDevicesTask(m_dsid));
    if (m_async)
    {
        dsa->addTask(boost::dynamic_pointer_cast<dSTask>(task));
    }
    else
    {
        task->run(dsa);
    }
}
StartupDSMTask::StartupDSMTask(dsid_t dsid, bool async) : dSTask(),
                                                    m_dsid(dsid), m_async(async)
{
    m_name = "dSM startup handler";
    m_description = "Check dSM configuration on startup";
    m_valid = true;
}

void StartupDSMTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    // 1. Check: Ping Settings
    // TODO: configure 24h ping mode every night at 3 a.m., through external
    // cron job?
    // TODO: is reboot required after setting ping parameters


    // 2. Check: Communication Log
    dsa->dS485()->setCommunicationLogging(m_dsid, 0xff);
}

SetDSMApartmentIdTask::SetDSMApartmentIdTask(dsid_t dsid, int apartment_id,
        bool async) :
        dSTask(), m_dsid(dsid), m_apartment_id(apartment_id), m_async(async)
{
    m_name = "Set apartment id handler";
    m_description = "Set dSM apartment id and restart";
    m_valid = true;
}

void SetDSMApartmentIdTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    boost::shared_ptr<std::list<dsid_t> > dsms;
    boost::shared_ptr<std::list<dsid_t> > dsm_list(new std::list<dsid_t>);
    if (IsNullId(m_dsid))
    {
        dsms = Storage::getInstance()->getDSMList();
    }
    else
    {
        dsms = dsm_list;
        dsms->push_back(m_dsid);
    }

    // loop over all dSMeter
    for (std::list<dsid_t>::iterator itm = dsms->begin();
         m_valid && (itm != dsms->end());
         itm++)
    {
        boost::shared_ptr<const dSM> dsmo =
                                        Storage::getInstance()->getDSM(*itm);
        boost::shared_ptr<dSM> dsm(dsmo->clone());

        uint8_t aid;
        dsa->dS485()->getApartmentId(*itm, &aid);

        DSID_LOG(logDEBUG2, *itm) << "[Task SetApartmentId] "
                "Database = " << util::str(m_apartment_id) <<
                ", Datamodel = " << util::str(dsm->getApartmentId()) <<
                ", Device = " << util::str(aid);

        if ((uint8_t) m_apartment_id != aid)
        {
            if (dsm->getApartmentId() >= 64)
            {
                FILE_LOG(logERROR) <<
                    "[Task SetApartmentId] Id mismatch on dsm " <<
                    util::str(*itm) <<
                    ": Database value " << util::str(m_apartment_id) <<
                    ", device value " << util::str(aid) <<
                    ", repeatedly failed to set id #" <<
                    util::str(dsm->getApartmentId() - 64);

                dsm->setApartmentId(dsm->getApartmentId() + 1);

                // TODO: Device Error Management
                // Populate misconfiguration error to dsm status monitor
                // CheckDSMTask
            }

            for (int retry = 0; retry < 5; retry ++)
            {
                DSID_LOG(logINFO, *itm) << "[Task SetApartmentId] Id "
                "reconfiguration from " << util::str(aid) <<
                " to " << util::str(m_apartment_id);

                aid = (uint8_t) m_apartment_id;
                try
                {
                    dsa->dS485()->setApartmentId(*itm, aid);
                }
                catch (Exception ex)
                {
                    DSID_LOG(logERROR, *itm) << "[Task SetApartmentId] "
                    "reconfiguration failed: " << ex.getMessage();
                    continue;
                }

                try
                {
                    dsa->dS485()->restartMeter(*itm);
                }
                catch(Exception ex)
                {
                    DSID_LOG(logERROR, *itm) << "[Task SetApartmentId] "
                    "restart failed: " << ex.getMessage();
                    continue;
                }
                break;
            }

            if (dsm->getApartmentId() >= 64) {
                dsm->setApartmentId(dsm->getApartmentId() + 1);
            }
            else {
                dsm->setApartmentId(64);
            }
            Storage::getInstance()->addDSM(dsm);
        }
    }
}

CheckDSMTask::CheckDSMTask(dsid_t dsid, bool async) : dSTask(), m_dsid(dsid),
        m_async(async)
{
    m_name = "dSM periodic health check handler";
    m_description = "Periodically check dSM state and configuration";
    m_valid = true;
}

void CheckDSMTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    // TODO:

    // 1. Query dSMState() and set global health status accordingly
    // 2. Calculate maximum 2busid = busmember + apartment id" and verify no
    //    dsm is above 62
    // 3. Set ping default mode: 24h (or from database?)
    // 4. Verify temperature limits (90/150 instead of 70/90)

}

GetActiveDevicesInZoneTask::GetActiveDevicesInZoneTask(dsid_t dsid,
                                                       uint16_t zone_id) :
                                                       dSTask(), m_dsid(dsid),
                                                       m_zone_id(zone_id)
{
    m_name = "Getting active devices from zone";
    m_description = "Retrieve information about active devices in a given zone"
                    " on a dS-Meter";
    m_valid = true;
}

void GetActiveDevicesInZoneTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    uint16_t count;

    uint16_t device_id;
    uint16_t vendor_id;
    uint16_t product_id;
    uint16_t function_id;
    uint16_t version;
    uint16_t zone_id;
    uint8_t state;
    uint8_t button_id;
    uint8_t locked;
    uint8_t output_mode;
    uint8_t groups[8];
    uint8_t name[21];
    uint8_t buttonGroups[16];
    uint8_t button_zone_membership[2];
    uint32_t serial_number;

    dsa->dS485()->getActiveDevicesCountInZone(m_dsid, m_zone_id, &count);
    for (uint16_t i = 0; i < count; i++)
    {
        dsa->dS485()->getActiveDeviceInfoInZone(m_dsid, m_zone_id, i,
                                       &device_id,
                                       &vendor_id, &product_id, &function_id,
                                       &version, &zone_id, &state, &button_id,
                                       &locked, &output_mode, groups,
                                       name, buttonGroups,
                                       button_zone_membership, &serial_number);

        dsid_t device_dsid;
        if (DsmApiExpandDeviceDSID(vendor_id, serial_number, &device_dsid) != 0)
        {
            FILE_LOG(logERROR) <<
                "[Task GetActiveDevices] Could not resolve dsid: "
                "serial " << util::hex(device_id) <<
                "/vendor " << util::hex(vendor_id);
            continue;
        }

        boost::shared_ptr<Device> dev(new Device(device_dsid, m_dsid));
        dev->setId(device_id);
        dev->setProductId(product_id);
        dev->setFunctionId(function_id);
        dev->setVersion(version);
        dev->setZoneId(zone_id);
        dev->setState(state);
        dev->setButtonId(button_id);
        dev->setLocked(locked);
        dev->setOutputMode(output_mode);
        dev->setName((const char*)name);
        dev->setSerialNumber(serial_number);
        Storage::getInstance()->addDevice(dev);

        FILE_LOG(logINFO) << "Added " << (state ? "active" : "inactive")
                          << " device with dSID " << util::str(dev->getDSID())
                          << " and address " << util::str(device_id)
                          << " on dSM " << util::str(dev->getDSMdSID());
    }
}

GetAllDevicesTask::GetAllDevicesTask(dsid_t dsid) :
        dSTask(), m_dsid(dsid)
{
    m_name = "Getting all devices on a circuit";
    m_description = "Retrieve information about all devices on a given circuit";
    m_valid = true;
}

void GetAllDevicesTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    uint16_t count;

    uint16_t device_id;
    uint16_t vendor_id;
    uint16_t product_id;
    uint16_t function_id;
    uint16_t version;
    uint16_t zone_id;
    uint8_t state;
    uint8_t button_id;
    uint8_t locked;
    uint8_t output_mode;
    uint8_t groups[8];
    uint8_t name[21];
    uint8_t buttonGroups[16];
    uint8_t button_zone_membership[2];
    uint32_t serial_number;

    dsa->dS485()->getDevicesCountInZone(m_dsid, 0, &count);
    for (uint16_t i = 0; i < count; i++)
    {
        try {
            dsa->dS485()->getDeviceInfoByIndex(m_dsid, 0, i,
                    &device_id, &vendor_id, &product_id, &function_id,
                    &version, &zone_id, &state, &button_id,
                    &locked, &output_mode, groups,
                    name, buttonGroups,
                    button_zone_membership, &serial_number);
        }
        catch (Exception& ex)
        {
            FILE_LOG(logERROR) << "dsm-api call exception: dSM:"
                               << util::str(m_dsid) << ", device index "
                               << util::str(i);
            continue;
        }

        dsid_t device_dsid;
        if (DsmApiExpandDeviceDSID(vendor_id, serial_number, &device_dsid) != 0)
        {
            FILE_LOG(logERROR) <<
                    "[Task GetAllDevices] Could not resolve dsid: "
                    "serial " << util::hex(device_id) <<
                    "/vendor " << util::hex(vendor_id);
            continue;
        }

        boost::shared_ptr<Device> dev(new Device(device_dsid, m_dsid));
        dev->setId(device_id);
        dev->setProductId(product_id);
        dev->setFunctionId(function_id);
        dev->setVersion(version);
        dev->setZoneId(zone_id);
        dev->setState(state);
        dev->setButtonId(button_id);
        dev->setLocked(locked);
        dev->setOutputMode(output_mode);
        dev->setName((const char*)name);
        dev->setSerialNumber(serial_number);

        //TODO: set zone index

        Storage::getInstance()->addDevice(dev);

        FILE_LOG(logINFO) << "Added " << (state ? "active" : "inactive")  <<
                " device with dSID " << util::str(dev->getDSID()) <<
                " and address " << util::str(device_id) <<
                " on dSM " << util::str(dev->getDSMdSID());
    }
}

GetDSMZonesTask::GetDSMZonesTask(dsid_t dsid) : dSTask(), m_dsid(dsid)
{
    m_name = "Getting number of zones on dSM";
    m_description = "Retrieve number of zones for dSM " + util::str(dsid);
    m_valid = true;
}

void GetDSMZonesTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    uint8_t count;
    dsa->dS485()->getZoneCount(m_dsid, &count);
    for (uint8_t i = 0; i < count; i++)
    {
        boost::shared_ptr<GetDSMZoneInfoTask> task(
                                            new GetDSMZoneInfoTask(m_dsid, i));
        dsa->addTask(boost::dynamic_pointer_cast<dSTask>(task));
    }
}

GetDSMZoneInfoTask::GetDSMZoneInfoTask(dsid_t dsid, uint8_t zone_index) :
                                       dSTask(), m_dsid(dsid),
                                       m_zone_index(zone_index)
{
    m_name = "Getting zone information";
    m_description = "Retrieve information for zone with index " +
                     util::str(zone_index) + " on dSM " + util::str(dsid);
    m_valid = true;
}

void GetDSMZoneInfoTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    uint16_t zone_id;
    uint8_t vzone_id;
    uint8_t num_grp;
    uint8_t name[21];

    dsa->dS485()->getZoneInfo(m_dsid, m_zone_index, &zone_id, &vzone_id,
                              &num_grp, name);
    printf("Retrieved zone info for dsm: %s, zoneid: %u vzid: %u groups: %u "
           "name %s\n", util::str(m_dsid).c_str(), zone_id, vzone_id,
           num_grp, name);
}

GetDeviceInfoTask::GetDeviceInfoTask(dsid_t dsm_dsid, uint16_t device_id) :
                                     dSTask(), m_dsid(dsm_dsid),
                                     m_device_id(device_id)
{
    m_name = "Getting device information";
    m_description = "Retrieve device information for device on dSM " +
                   util::str(dsm_dsid);
    m_valid = true;
}

void GetDeviceInfoTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        return;
    }

    uint16_t device_id;
    uint16_t vendor_id;
    uint16_t product_id;
    uint16_t function_id;
    uint16_t version;
    uint16_t zone_id;
    uint8_t state;
    uint8_t button_id;
    uint8_t locked;
    uint8_t output_mode;
    uint8_t groups[8];
    uint8_t name[21];
    uint8_t buttonGroups[16];
    uint8_t button_zone_membership[2];
    uint32_t serial_number;

    dsa->dS485()->getDeviceInfoById(m_dsid, m_device_id,
                       &device_id, &vendor_id, &product_id, &function_id,
                       &version, &zone_id, &state, &button_id, &locked,
                       &output_mode, groups, name, buttonGroups,
                       button_zone_membership, &serial_number);

    dsid_t device_dsid;
    if (DsmApiExpandDeviceDSID(vendor_id, serial_number, &device_dsid) != 0)
    {
        printf("Could not resolve dsid: vendor id is not known, ignoring "
               "device");
        return;
    }

    boost::shared_ptr<const Device> dsdo =
        Storage::getInstance()->getDevice(device_dsid, m_dsid);
    boost::shared_ptr<Device> dev(
            dsdo == NULL ? new Device(device_dsid, m_dsid) : dsdo->clone());

    dev->setId(device_id);
    dev->setProductId(product_id);
    dev->setFunctionId(function_id);
    dev->setVersion(version);
    dev->setZoneId(zone_id);
    dev->setState(state);
    dev->setButtonId(button_id);
    dev->setLocked(locked);
    dev->setOutputMode(output_mode);
    dev->setName((const char*)name);
    Storage::getInstance()->addDevice(dev);

    FILE_LOG(logINFO) << "Added device with dSID "
                      << util::str(dev->getDSID()) <<  " on dSM "
                      << util::str(dev->getDSMdSID());
}

TestTranmissionQualityTask::TestTranmissionQualityTask(int checksPerDevice,
                                dsid_t dsm_dsid) : dSLongTask(),
                                m_checksPerDevice(checksPerDevice),
                                m_dsm_dsid(dsm_dsid)
{
    m_name = "Test der Übertragungsqualität der Klemmen";
    m_description = "Testen der Übertragungsqualität für ";
    m_parameters = "dsm=" + util::str(dsm_dsid) + '&' +
                   "checksPerDevice=" + util::str(checksPerDevice);

    if (IsNullId(dsm_dsid))
    {
        m_description += "alle Klemmen";
    }
    else
    {
        m_description += "alle Klemmen am dSM " + util::str(dsm_dsid);
    }

    m_description += ", Tests pro Klemme: " + util::str(checksPerDevice);

    m_valid = true;
}

task_type TestTranmissionQualityTask::getTaskType() const
{
    return TestTransmissionQuality;
}

void TestTranmissionQualityTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_status_mutex.unlock();

    boost::shared_ptr<std::list<boost::shared_ptr<const Device> > > devs;
    if (IsNullId(m_dsm_dsid))
    {
        devs = Storage::getInstance()->getDevices(true);
    }
    else
    {
        devs = Storage::getInstance()->getDevices(m_dsm_dsid, true);
    }

    size_t total = devs->size() * m_checksPerDevice;
    size_t current = 0;
    int check_loop = 0;

    if (0 == total)
    {
        m_status_mutex.lock();
        m_progress = 100;
        m_task_status = Finished;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_progress = (unsigned char)(current * 100 / total);
    m_status_mutex.unlock();

        std::list<boost::shared_ptr<const Device> >::const_iterator it =
                                                                devs->begin();

    while (m_valid && (current < total))
    {
        boost::shared_ptr<const Device> device = *it;

        current++;

        m_status_mutex.lock();
        m_status_info = "Requesting transmission quality from device " +
                        util::str(device->getDSID());
        if (m_checksPerDevice > 1)
        {
            m_status_info += ", test " + util::str(check_loop + 1);
        }
        m_status_mutex.unlock();

        uint8_t down = TRANSMISSION_QUALITY_DOWNSTREAM_MAX;
        uint16_t up = 0;
        int err = 0;
        try
        {
            dsa->dS485()->testDeviceTransmissionQuality(device->getDSMdSID(),
                                                   device->getId(), &down, &up);

            FILE_LOG(logINFO) << "Transmission quality for device " <<
                    util::str(device->getDSID()) <<
                    ", up " << util::str(up) << ", down " << util::str(down);
        }
        catch (Exception ex)
        {
            m_status_mutex.lock();
            m_status_info = "Failed to get transmission quality for "
                            "device " + util::str(device->getDSID()) +
                            ": " + ex.getMessage();
            err = ex.getErrorCode();
            m_status_mutex.unlock();

                FILE_LOG(logDEBUG1) << "[Task Transmission Quality] device " <<
                            util::str(device->getDSID()) <<
                            ", error "  << util::str(err) << " = "
                            << ex.getMessage();
        }

        Database::getInstance()->saveTransmissionQuality(m_task_id,
                                util::str(device->getDSID()), up, down, err);

        m_status_mutex.lock();
        m_progress = (unsigned char)(current * 100 / total);
        m_status_mutex.unlock();

        it++;
        if (it == devs->end())
        {
            check_loop++;
            if (check_loop < m_checksPerDevice)
            {
                it = devs->begin();
            }
            else
            {
                break;
            }
        }
    }

    m_status_mutex.lock();
    if (!m_valid)
    {
        m_task_status = Cancelled;
        m_status_info = "Cancelled";
    }
    else
    {
        m_task_status = Finished;
        m_status_info = "Finished";
    }
    m_status_mutex.unlock();
}

TestDevicePresenceTask::TestDevicePresenceTask(dsid_t dsm_dsid,
                                               int interval_time) :
        dSLongTask(),
        m_dsm_dsid(dsm_dsid),
        m_interval_time(interval_time),
        m_devs(new std::list<boost::shared_ptr<DeviceP> >),
        m_dsms(new std::list<boost::shared_ptr<MeterP> >)
{
    m_name = "Test device presence";
    m_description = "Check device status changes for ";
    m_parameters = "dsm=" + util::str(dsm_dsid) + '&' +
            "interval=" + util::str(interval_time);

    if (IsNullId(dsm_dsid))
    {
        m_description += "all devices on dSM " + util::str(dsm_dsid);
    }
    else
    {
        m_description += "all devices";
    }
    m_description += ", interval time: " + util::str(m_interval_time) + "ms";
    m_valid = true;

    pthread_mutex_init(&m_ping_lock, NULL);
    pthread_cond_init(&m_ping_status_change, NULL);
}

TestDevicePresenceTask::~TestDevicePresenceTask()
{
    pthread_mutex_destroy(&m_ping_lock);
    pthread_cond_destroy(&m_ping_status_change);
}

task_type TestDevicePresenceTask::getTaskType() const
{
    return TestDevicePresence;
}

TestDevicePresenceTask::DeviceP::DeviceP(
                    boost::shared_ptr<const Device> &parent) : Device(*parent)
{
}

TestDevicePresenceTask::MeterP::MeterP(boost::shared_ptr<const dSM> &parent,
        TestDevicePresenceTask* const task) :
        dSM(*parent),
        m_task(task)
{
}

void TestDevicePresenceTask::MeterP::resetPingState()
{
    m_count_ping_dut = m_count_ping_sent = m_count_ping_received = m_deviceid_dut = 0;
    m_state = unknown;
}

void TestDevicePresenceTask::MeterP::getPingStatus(uint16_t* tx_dev_count,
                                                   uint16_t* tx_count,
                                                   uint16_t* rx_count)
{
    *tx_dev_count = m_count_ping_dut;
    *tx_count = m_count_ping_sent;
    *rx_count = m_count_ping_received;
}

void TestDevicePresenceTask::MeterP::signalPingSent(uint16_t device_id)
{
    if (device_id < 64)
    {
        return;
    }
    if (device_id != m_deviceid_dut)
    {
        m_deviceid_dut = device_id;
        m_count_ping_dut ++;
        m_count_ping_timeout = 0;
    }
    m_state = ping_sent;
    m_count_ping_sent ++;
    pthread_cond_signal(&m_task->m_ping_status_change);
}

void TestDevicePresenceTask::MeterP::signalPingReply(uint16_t device_id)
{
    if (device_id == m_deviceid_dut)
    {
        m_count_ping_received ++;
        m_state = reply_received;
        if (m_count_ping_dut >= getDeviceCount())
        {
            m_state = ready;
        }
    }
    pthread_cond_signal(&m_task->m_ping_status_change);
}

void TestDevicePresenceTask::MeterP::signalPingTimeout()
{
    m_state = waiting;
    if (++m_count_ping_timeout >= 3)
    {
        if (m_count_ping_dut >= getDeviceCount())
        {
            m_state = ready;
        }
    }
}

boost::shared_ptr<TestDevicePresenceTask::MeterP>
                                    TestDevicePresenceTask::getDSM(dsid_t dsmid)
{
    for (std::list<boost::shared_ptr<
            TestDevicePresenceTask::MeterP> >::iterator itm = m_dsms->begin();
            (itm != m_dsms->end());
            itm++)
    {
        dsid_t id = (itm->get())->getDSID();
        if (IsEqualId(id, dsmid))
        return *itm;
    }
    return boost::shared_ptr<TestDevicePresenceTask::MeterP>();
}

boost::shared_ptr<TestDevicePresenceTask::DeviceP>
            TestDevicePresenceTask::getDevice(dsid_t dsmid, uint16_t devindex)
{
    for (std::list<boost::shared_ptr<
            TestDevicePresenceTask::DeviceP> >::iterator itm = m_devs->begin();
            (itm != m_devs->end());
            itm++)
    {
        dsid_t idm = (*itm)->getDSMdSID();
        if (IsEqualId(idm, dsmid) && ((*itm)->getId() == devindex))
            return *itm;
    }
    return boost::shared_ptr<TestDevicePresenceTask::DeviceP>();
}

boost::shared_ptr<TestDevicePresenceTask::DeviceP>
                TestDevicePresenceTask::getDevice(dsid_t dsmid, dsid_t devid)
{
    for (std::list<boost::shared_ptr<
            TestDevicePresenceTask::DeviceP> >::iterator itm = m_devs->begin();
            (itm != m_devs->end());
            itm++)
    {
        dsid_t idm = (itm->get())->getDSMdSID();
        dsid_t idd = (itm->get())->getDSID();
        if (IsEqualId(idm, dsmid) && IsEqualId(idd, devid))
            return *itm;
    }
    return boost::shared_ptr<TestDevicePresenceTask::DeviceP>();
}

void TestDevicePresenceTask::callback(boost::shared_ptr<CBData> busdata)
{
    switch(busdata->getType())
    {

        default:
            break;

        case CommUpstreamShort:
            break;

        case CommUpstreamLong:
        {
            boost::shared_ptr<CommUpstreamLongData> data =
                boost::dynamic_pointer_cast<CommUpstreamLongData> (busdata);
            boost::shared_ptr<MeterP> dsm = getDSM(data->getDSMSourceDSID());
            switch (data->getMainQualifier())
            {
                case 15:
                {
                    // NOTE: this 32bit -> DSID conversion assumes
                    // vendor id 0x0001
                    dsid_t device_dsid;
                    DsmApiExpandDeviceDSID(0x0001, data->getData(),
                                           &device_dsid);
                    boost::shared_ptr<DeviceP> dsd =
                            getDevice(data->getDSMSourceDSID(), device_dsid);
                    if (dsd && dsd->getPingState() == sent)
                    {
                        FILE_LOG(logDEBUG2) << "Ping reply from "
                                            << util::str(dsd->getId()) << "/"
                                            << util::hex(data->getData());

                        switch(dsd->getPingState())
                        {
                            case sent:
                                dsd->setPingState(replied);
                                dsm->signalPingReply(dsd->getId());
                                break;
                            case replied:
                                break;
                            default:
                                break;
                        }
                    }
                }
                break;
            }
        }
        break;

        case CommDownstreamShort:
        break;

        case CommDownstreamLong0:
        {
            boost::shared_ptr<CommDownstreamLong0Data> data(
                boost::dynamic_pointer_cast<CommDownstreamLong0Data> (busdata));
            boost::shared_ptr<MeterP> dsm = getDSM(data->getDSMSourceDSID());
            boost::shared_ptr<DeviceP> dsd;
            if (dsm)
            {
                switch (data->getCommand())
                {
                    case 0x02:
                    {
                        dsd = getDevice(data->getDSMSourceDSID(),
                                                    dsm->getSelectionB() >> 6);
                        if (dsd)
                        {
                            if (dsd->getPingState() == scheduled)
                            {
                                FILE_LOG(logDEBUG2) << "Ping sent to " <<
                                util::str(dsm->getSelectionB() >> 6) << "/" <<
                                util::str(dsd->getDSID());
                                dsd->setPingState(sent);
                                dsm->signalPingSent(dsm->getSelectionB() >> 6);
                            }
                            else if (dsd->getPingState() == sent)
                            {
                                FILE_LOG(logDEBUG2)
                                        << "Ping repeatedly sent to "
                                        << util::str(dsm->getSelectionB() >> 6)
                                        << "/" << util::str(dsd->getDSID());
                                dsm->signalPingSent(dsm->getSelectionB() >> 6);
                            }
                        }
                    } /* case 0x02 */
                    break;
                } /* switch getCommand() */
            } /* if (dsm) */
        }
        break;

        case CommDownstreamLong8:
        break;

        case CommDownstreamLong16:
        {
            boost::shared_ptr<CommDownstreamLong16Data> data(
                boost::dynamic_pointer_cast<CommDownstreamLong16Data>(busdata));
            boost::shared_ptr<MeterP> dsm = getDSM(data->getDSMSourceDSID());
            if (dsm)
            {
                static uint16_t dsidh = 0;
                uint16_t payload = data->getData();
                switch (data->getCommand())
                {
                    case 0:
                        dsm->setSelectionA(payload);
                        break;
                    case 1:
                        dsm->setSelectionB(payload);
                        dsm->setSelectionD(0);
                        break;
                    case 2:
                        dsidh = payload;
                        break;
                    case 3:
                        dsm->setSelectionD(
                                (uint32_t) dsidh << 16 | data->getData());
                        dsidh = 0;
                        break;
                }
            }
        }
        break;

        case DeviceAdded:
        {
            boost::shared_ptr<DeviceAddedData> data(
                    boost::dynamic_pointer_cast<DeviceAddedData> (busdata));
            boost::shared_ptr<DeviceP> dsd = getDevice(data->getDSMSourceDSID(),
                                                       data->getDeviceID());
            if (dsd)
            {
                dsd->setPingState(discovered);
            }
        }
        break;

        case DeviceRemoved:
        {
            boost::shared_ptr<DeviceRemovedData> data(
                    boost::dynamic_pointer_cast<DeviceRemovedData> (busdata));
            boost::shared_ptr<DeviceP> dsd = getDevice(data->getDSMSourceDSID(),
                                                       data->getDeviceID());
            if (dsd)
            {
                dsd->setPingState(removed);
            }
        }
        break;
    }
}

void TestDevicePresenceTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_progress = 0;
    m_status_mutex.unlock();

    boost::shared_ptr<std::list<boost::shared_ptr<const Device> > > devs;
    boost::shared_ptr<std::list<dsid_t> > dsms;
    boost::shared_ptr<std::list<dsid_t> > dsm_list(new std::list<dsid_t>);

    if (IsNullId(m_dsm_dsid))
    {
        devs = Storage::getInstance()->getDevices(false);
        dsms = Storage::getInstance()->getDSMList();
    }
    else
    {
        devs = Storage::getInstance()->getDevices(m_dsm_dsid, false);
        dsms = dsm_list;
        dsms->push_back(m_dsm_dsid);
    }

    size_t total_tx = 0;
    size_t total_rx = 0;
    size_t total_devs = devs->size();

    if (0 == total_devs)
    {
        m_status_mutex.lock();
        m_progress = 100;
        m_task_status = Finished;
        m_status_mutex.unlock();
        return;
    }

    FILE_LOG(logINFO) << "[Task Device Presence] Starting presence check on " <<
            util::str(dsms->size()) << " meter(s) for a total of " <<
            util::str(total_devs) << " devices";

    // disable all active pings
    dsid_t bid;
    SetBroadcastId(bid);
    try
    {
        dsa->dS485()->pingOff(bid);
    }
    catch (Exception ex)
    {
        // Broadcast command -> result check raises exception for return code 2 :(
    }

    // copy device list and preset device ping state
    for (std::list<boost::shared_ptr<const Device> >::iterator it =
            devs->begin();
            it != devs->end();
            it++)
    {
        boost::shared_ptr<DeviceP> dsd(new DeviceP(*it));
        dsd->setPingState(scheduled);
        m_devs->push_back(dsd);
    }

    // copy meter list and reset ping state
    for (std::list<dsid_t>::iterator itm = dsms->begin();
            m_valid && (itm != dsms->end());
            itm++)
    {
        boost::shared_ptr<const dSM> dsmo =
                                        Storage::getInstance()->getDSM(*itm);
        boost::shared_ptr<MeterP> dsm(new MeterP(dsmo, this));
        dsm->resetPingState();
        m_dsms->push_back(dsm);
    }

    // add task to receive callbacks
    dsa->dS485()->addListener(this);

    // select the callbacks you want
    this->select(CommUpstreamShort, true);
    this->select(CommUpstreamLong, true);
    this->select(CommDownstreamShort, true);
    this->select(CommDownstreamLong0, true);
    this->select(CommDownstreamLong8, true);
    this->select(CommDownstreamLong16, true);
    this->select(DeviceAdded, true);
    this->select(DeviceRemoved, true);

    // loop over all dSMeter
    for (std::list<boost::shared_ptr<
            TestDevicePresenceTask::MeterP> >::iterator itm = m_dsms->begin();
            m_valid && (itm != m_dsms->end());
            itm++)
    {
        uint16_t dev_count = (*itm)->getDeviceCount();
        uint16_t tx_dev_count = 0;
        uint16_t tx_count = 0;
        uint16_t rx_count = 0;
        uint16_t tmo_auto = 20000;
        uint16_t tmo_device = 60000;
        int rc = 0;
        int wait;
        struct timespec ts;
        struct timeval now;

        m_status_mutex.lock();
        m_status_info = "Check " + util::str(dev_count) + " devices on dsm " +
                util::str((*itm)->getDSID());
        m_status_mutex.unlock();

        if (dev_count == 0)
        {
            continue;
        }

        {
            uint8_t Hardware;
            uint8_t DSP;
            uint8_t Temperature;
            uint8_t Config;
            uint8_t dS485;
            uint8_t Line;
            dS_Assistant::getInstance()->dS485()->dSMState((*itm)->getDSID(),
                    &Hardware, &DSP, &Temperature, &Config, &dS485, &Line);
            if (Line)
            {
                FILE_LOG(logINFO) << "[Task Device Presence] dSM " <<
                        util::str((*itm)->getDSID()) << " is currently jumbling "
                        "and busy, device presence check skipped";
                total_tx += dev_count;
                continue;
            }
        }

        FILE_LOG(logDEBUG1) << "[Task Device Presence] Check " <<
                util::str(dev_count) << " devices on dsm " <<
                util::str((*itm)->getDSID());

        // send ping-one-time-off command, after that the 24h mode is re-enabled
        try
        {
            dsa->dS485()->getPingIntervals((*itm)->getDSID(), &tmo_auto,
                                           &tmo_device);
            dsa->dS485()->pingOneTimeOff((*itm)->getDSID(),
                                         (uint16_t) m_interval_time);
        }
        catch(Exception ex)
        {
            FILE_LOG(logERROR)
                << "[Task Device Presence] dSM-API request failed: "
                << ex.getMessage();
            continue;
        }

        FILE_LOG(logDEBUG4) << "[Task Device Presence] Configured ping times, "
                "dsm-auto-delay " << util::str(tmo_auto) << ", "
                "dsm-auto-interval " << util::str(tmo_device) << ", "
                "request-interval " << util::str(m_interval_time);

        while (m_valid)
        {
            wait = 5000;
            switch((*itm)->getState())
            {
                case waiting:
                case unknown:
                case reply_received:
                    // waiting for next downstream ping
                    wait = 2500 + m_interval_time;
                    break;

                case ping_sent:
                    // waiting for upstream response
                    wait = 2500;
                    break;

                default:
                    break;
            }

            gettimeofday(&now, NULL);
            ts.tv_sec = now.tv_sec + wait / 1000;
            ts.tv_nsec = now.tv_usec * 1000 + (wait % 1000) * 1000;

            pthread_mutex_lock(&m_ping_lock);
            rc = pthread_cond_timedwait(&m_ping_status_change, &m_ping_lock, &ts);
            (*itm)->getPingStatus(&tx_dev_count, &tx_count, &rx_count);
            pthread_mutex_unlock(&m_ping_lock);

            switch((*itm)->getState())
            {
                case unknown:
                case waiting:
                    break;

                case reply_received:
                    // as of maximum three checks per device add additional progress
                    // indication for "normal" replies
                    (*itm)->setState(waiting);
                    // fall through

                case ready:
                    total_tx += 2 - (*itm)->getPingRetries();
                    break;

                case ping_sent:
                    if (rc == ETIMEDOUT)
                    {
                        // ping reception timeout
                        (*itm)->signalPingTimeout();
                    }
                    break;

                default:
                    break;
            }

            m_status_mutex.lock();
            m_status_info = "Check " + util::str(dev_count) +
                " devices on dsm " + util::str((*itm)->getDSID()) +
                ", " + util::str(tx_count) + " completed";
            m_progress = (unsigned char)(
                                (total_tx + tx_count) * 100 / (3 * total_devs));
            m_status_mutex.unlock();

            FILE_LOG(logDEBUG4) << "[Task Device Presence] Ping Progress = "
                    << util::str(m_progress) <<
                    ", state = " << (*itm)->getState() <<
                    ", tx-total = " << util::str(total_tx) <<
                    ", tx = " << util::str(tx_count) <<
                    ", tx-dev = " << util::str(tx_dev_count) <<
                    ", rx = " << util::str(rx_count);

            boost::shared_ptr<const dSM> dsmo =
                            Storage::getInstance()->getDSM((*itm)->getDSID());
            if (! dsmo->getPresent())
            {
                FILE_LOG(logERROR)
                    << "[Task Device Presence] dSM inactive, check aborted";
                break;
            }
            if ((*itm)->pingFinished() || (rx_count >= dev_count))
            {
                break;
            }
        }

        total_tx += tx_count;
        total_rx += rx_count;
    }

    // don't forget to unsubscribe from the listener
    dsa->dS485()->removeListener(this);

    // loop over devices and count successful replies
    size_t total_replies = 0;
    size_t total_removed = 0;
    size_t total_discovered = 0;
    for (std::list<boost::shared_ptr<
            TestDevicePresenceTask::DeviceP> >::iterator it = m_devs->begin();
            it != m_devs->end();
            it++)
    {
        switch ((*it)->getPingState())
        {
            case replied:
                total_replies ++;
                break;
            case removed:
                total_removed ++;
                break;
            case discovered:
                total_discovered ++;
                break;
            default:
                break;
        }
    }

    // TODO: store in database

    FILE_LOG(logINFO) << "[Task Device Presence] Finished, "
            "total " << util::str(total_devs) << " devices checked "
            "and " << util::str(total_replies) << " are active now, " <<
            util::str(total_removed) << " devices changed to inactive and " <<
            util::str(total_discovered) << " to active";

    m_status_mutex.lock();
    if (!m_valid)
    {
        m_task_status = Cancelled;
        m_status_info = "Cancelled";
    }
    else
    {
        m_task_status = Finished;
        m_status_info = "Finished";
    }
    m_status_mutex.unlock();
}

TestDeviceStatisticTask::TestDeviceStatisticTask(dsid_t dsm_dsid,
        dsid_t dev_dsid) :
        dSLongTask(),
        m_dsm_dsid(dsm_dsid),
        m_dev_dsid(dev_dsid)
{
    m_name = "Statistik Daten des Gerätes";
    m_description = "Abfrage der Statistikdaten von " + util::str(dev_dsid);
    m_parameters = "dsm=" + util::str(dsm_dsid) + "&dsid=" +
                                                        util::str(dev_dsid);
    m_valid = true;
}

task_type TestDeviceStatisticTask::getTaskType() const
{
    return TestDeviceStatistic;
}

void TestDeviceStatisticTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_progress = 0;
    m_status_mutex.unlock();

    FILE_LOG(logINFO) << "[Task Device Statistic] Started query from device " <<
            util::str(m_dev_dsid) << " on dsm " << util::str(m_dsm_dsid);

    boost::shared_ptr<const Device> dsdo =
            Storage::getInstance()->getDevice(m_dev_dsid, m_dsm_dsid);
    if (NULL == dsdo)
    {
        m_status_info = "Ungültige Parameter";
        m_task_status = Finished;
        return;
    }

    size_t total = 0x29;
    size_t current = 0;
    size_t valid = 0;
    boost::shared_ptr<StatisticDevice> pstats;

    while (m_valid && (current < total))
    {
        m_status_mutex.lock();
        m_progress = (unsigned char)(current * 100 / total);
        m_status_info = "Abfrage vom Gerät: " + util::str(current) +
            " von " + util::str(total);
        m_status_mutex.unlock();

        int err = 0;
        int retry;
        uint16_t value;

        for (retry = 3; retry > 0; retry --)
        {
            try
            {
                dsa->dS485()->getStatisticValue(dsdo->getDSMdSID(),
                        dsdo->getId(), current, &value);
                valid ++;
                break;
            }
            catch (Exception ex)
            {
                m_status_mutex.lock();
                m_status_info = "Abfrage der Statistik Werte von "
                        "dSID " + util::str(dsdo->getDSID()) +
                        ", Datenindex " + util::str(current) +
                        ": " + ex.getMessage();
                    err = ex.getErrorCode();
                m_status_mutex.unlock();

                FILE_LOG(logDEBUG1) << "[Task Device Statistic] device " <<
                util::str(dsdo->getDSID()) <<
                ", index " << util::str(current) <<
                ", error " << util::str(err) << " = " << ex.getMessage();
            }
        }

        if (current == 0)
        {
            if (retry == 0)
            {
                m_status_mutex.lock();
                m_task_status = Finished;
                m_progress = 0;
                m_status_info =
                    "Fehler bei der Abfrage des ersten Statistikwertes";
                m_status_mutex.unlock();
                return;
            }
            total = (uint8_t) (value >> 8) / 2;
            pstats = boost::shared_ptr<StatisticDevice> (
                                            new StatisticDevice(value & 0xff));
        }

        if (retry)
        {
            pstats->pushValue(value);
        }
        else
        {
            pstats->pushNilValue(2);
        }

        current++;
    }

    FILE_LOG(logINFO) << "[Task Device Statistic] Query complete, got " <<
            util::str(current) << " values from device " <<
            util::str(dsdo->getDSID());
    if (current != valid)
    {
        FILE_LOG(logERROR) << "[Task Device Statistic] "
                "Errors occurred: " <<
                util::str(total-valid) << " values missing";
    }

    DSID_PREFIX_LOG(logINFO, dsdo->getDSID(), "stat_")
        << "Statistic Query Results:";
    boost::shared_ptr<std::list<std::string> > result = pstats->genResultList();
    for (std::list<std::string>::iterator it = result->begin();
            it != result->end();
            it++)
    {
        DSID_PREFIX_LOG(logINFO, dsdo->getDSID(), "stat_") << *it;
    }

    // TODO: store in database

    m_status_mutex.lock();
    if (!m_valid)
    {
        m_task_status = Cancelled;
        m_status_info = "Cancelled";
    }
    else
    {
        m_task_status = Finished;
        m_status_info = "Finished";
    }
    m_status_mutex.unlock();
}

TestMeterStatisticTask::TestMeterStatisticTask(dsid_t dsm_dsid) :
        dSLongTask(), m_dsm_dsid(dsm_dsid)
{
    m_name = "Statistik Daten des Meters";
    m_description = "Abfrage der Statistikdaten von dSM " +
        util::str(m_dsm_dsid);
    m_parameters = "dsm=" + util::str(m_dsm_dsid);
    m_valid = true;
}

task_type TestMeterStatisticTask::getTaskType() const
{
    return TestMeterStatistic;
}

void TestMeterStatisticTask::run(dS_Assistant *dsa)
{
    if (!(dsa && m_valid))
    {
        m_status_mutex.lock();
        m_task_status = Cancelled;
        m_status_mutex.unlock();
        return;
    }

    m_status_mutex.lock();
    m_task_status = Running;
    m_status_mutex.unlock();

    boost::shared_ptr<std::list<dsid_t> > dsms;
    boost::shared_ptr<std::list<dsid_t> > dsm_list(new std::list<dsid_t>);

    if (IsNullId(m_dsm_dsid))
    {
        dsms = Storage::getInstance()->getDSMList();
    }
    else
    {
        dsms = dsm_list;
        dsms->push_back(m_dsm_dsid);
    }

    size_t total = dsms->size();
    size_t current = 0;
    if (total == 0)
    {
        m_status_mutex.lock();
        m_progress = 100;
        m_task_status = Finished;
        m_status_mutex.unlock();
        return;
    }

    // loop over all dSMeter
    for (std::list<dsid_t>::iterator itm = dsms->begin();
            m_valid && (itm != dsms->end());
            itm++)
    {
        current ++;
        m_status_mutex.lock();
        m_progress = (unsigned char)(current * 100 / total);
        m_status_info = "Abfrage vom Gerät: " + util::str(current) +
            " von " + util::str(total);
        m_status_mutex.unlock();

        FILE_LOG(logINFO) << "[Task Meter Statistic] Started query from dsm " <<
                util::str(*itm);

        uint8_t length;
        uint8_t index;
        uint8_t data[256];
        try
        {
            length = index = 0;
            dsa->dS485()->getStatisticUpstream(*itm, 0, &length, data + index);
            index += length;
            dsa->dS485()->getStatisticUpstream(*itm, 1, &length, data + index);
            index += length;
        }
        catch (Exception& ex)
        {
            FILE_LOG(logERROR) << "[Task Meter Statistic] Query from dsm " <<
                    util::str(*itm) << " failed: " << ex.getMessage();
            continue;
        }
        uint16_t sversion = *(uint16_t *) data;
        boost::shared_ptr<StatisticMeter> pstats(new StatisticMeter(sversion));

        for (int i = 0; i < index; i++)
        {
            pstats->pushValue(data[i]);
        }

        boost::shared_ptr<std::list<std::string> > result =
            pstats->genResultList();
        for (std::list<std::string>::iterator it = result->begin();
                it != result->end();
                it++)
        {
            DSID_PREFIX_LOG(logINFO, *itm, "stat_") << *it;
        }

        // TODO: store in database
    }

    m_status_mutex.lock();
    if (!m_valid)
    {
        m_task_status = Cancelled;
        m_status_info = "Cancelled";
    }
    else
    {
        m_task_status = Finished;
        m_status_info = "Finished";
    }
    m_status_mutex.unlock();
}

} // namespace
