/*
    Copyright (c) 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

*/

#ifdef HAVE_CONFIG_H
#include "autoconfig.h"
#endif

#include <stdio.h>
#include <sstream>

#include "exceptions.h"
#include "database.h"
#include "util.h"
#include "dsmapi-wrapper.h"
#include "html_report_web_handler.h"

#define MAX_FETCH_RECORDS                   50

namespace dSA
{

HTMLReportWebHandler::HTMLReportWebHandler(struct mg_connection* connection,
                               const struct mg_request_info* request_info) :
                               HTMLWebHandler(connection, request_info) {}


std::string HTMLReportWebHandler::taskOverview(
                                            boost::shared_ptr<dSLongTask> task)
{
    std::ostringstream result;

    result << header("Übersicht");

    result << table_begin();

    result << table_row_begin();
    result << table_column_add(task->getName(), 2);
    result << table_row_end();

    result << table_row_begin();
    result << table_column_add(task->getDescription() + "<br/>", 2);
    result << table_row_end();

    result << table_row_begin();
    result << table_column_add("Status:", 0, 250);
    result << table_column_add(task->getStatusInfo());
    result << table_row_end();

    task_status t = task->getTaskStatus();
    if (t == Queued)
    {
        result << table_end();
        return result.str();
    }

    result << table_row_simple("Gestartet:",
                            util::epoch2local((time_t)task->getStartedTime()));

    if ((t != Cancelled) && (t != Running))
    {
        result << table_row_simple("Beendet:",
                            util::epoch2local((time_t)task->getFinishedTime()));
    }

    if (t == Running)
    {
        result << table_row_begin();
        result << table_column_add("<b>Hinweis:</b> ein Protokoll kann nur "
                                   "generiert werden, wenn der Test beendet "
                                   "ist.", 2);
        result << table_row_end();

        result << table_end();
        return result.str();
    }


    result << table_end();
    return result.str();
}

std::string HTMLReportWebHandler::upstreamPercent(int up)
{
    return util::str(up * 100 / TRANSMISSION_QUALITY_UPSTREAM_MAX) + '%';
}

std::string HTMLReportWebHandler::downstreamPercent(int down)
{
    return util::str(100 - (down * 100 / TRANSMISSION_QUALITY_DOWNSTREAM_MAX)) +
           '%';
}

std::string HTMLReportWebHandler::transmissionQualitySummary(long long int id)
{
    std::ostringstream result;

    boost::shared_ptr<Database::task_result_summary> summary =
        Database::getInstance()->getTaskResultSummary(id);

    boost::shared_ptr<Database::result_data> data = summary->data;

    result << header("Zusammenfassung");
    result << table_begin();

    result << table_row_begin();
    result << table_column_add("Anzahl der getesteten Geräte:", 0, 250);
    result << table_column_add((*data)["devices"]->first);
    result << table_row_end();

    result << table_row_simple("Anzahl der durchgeführten Tests:",
                               (*data)["tests"]->first);

     result << table_row_simple("Anzahl der Tests mit Fehlern:",
                               (*data)["errors"]->first);

    result << table_row_begin();
    result << table_column_add("&nbsp;", 2);
    result << table_row_end();


    result << table_row_simple("Durchschnittliche Rückkanal Qualität:",
                upstreamPercent(util::str2int((*data)["avgUpstream"]->first)));

    result << table_row_simple("Maximale Rückkanal Qualität:",
                upstreamPercent(util::str2int((*data)["maxUpstream"]->first)));

    result << table_row_simple("Minimale Rückkanal Qualität:",
                upstreamPercent(util::str2int((*data)["minUpstream"]->first)));

    result << table_row_begin();
    result << table_column_add("&nbsp;", 2);
    result << table_row_end();

    result << table_row_simple("Durchschnittliche Hinkanal Qualität:",
            downstreamPercent(util::str2int((*data)["avgDownstream"]->first)));

    result << table_row_simple("Maximale Hinkanal Qualität:",
            downstreamPercent(util::str2int((*data)["maxDownstream"]->first)));

    result << table_row_simple("Minimale Hinkanal Qualität:",
            downstreamPercent(util::str2int((*data)["minDownstream"]->first)));

    result << table_end();
    return result.str();
}

std::string HTMLReportWebHandler::renderTQDevREcord(
        boost::shared_ptr<
            const Database::transmission_quality_device_record> rec)
{
    std::ostringstream result;

    result << table_row_begin();
    result << table_column_add(rec->dsid);
    if (rec->errorcode == 0)
    {
        result << table_column_add(upstreamPercent(rec->upstream));
        result << table_column_add(downstreamPercent(rec->downstream));
    }
    else
    {
        result << table_column_add("-");
        result << table_column_add("-");
    }
    result << table_column_add(dSMApiWrapper::error2Message(rec->errorcode));
    result << table_column_add(util::epoch2local(rec->timestamp));
    result << table_row_end();
    return result.str();
}

void HTMLReportWebHandler::transmissionQualityReport(
                                            boost::shared_ptr<dSLongTask> task,
                                            bool download)
{
    long long int c = 0;
    std::string s = page_begin("Protokoll Übertragungsqualität", download);
    mg_write(m_connection, s.c_str(), s.length());

    try
    {
        s = taskOverview(task);
        mg_write(m_connection, s.c_str(), s.length());

        task_status t = task->getTaskStatus();
        if ((t != Finished) && (t != Cancelled))
        {
            s = page_end();
            mg_write(m_connection, s.c_str(), s.length());
            return;
        }

        s = transmissionQualitySummary(task->getTaskID());
        mg_write(m_connection, s.c_str(), s.length());

        Database* db = Database::getInstance();
        c = db->transmissionQualityResultCount(task->getTaskID());
        long long int offset = 0;

        std::ostringstream html;

        if (c > 0)
        {
            s = header("Detaillierte Ergebnisse");
            mg_write(m_connection, s.c_str(), s.length());

            html << table_begin();

            html << table_row_begin();
            html << table_header("dSID");
            html << table_header("Rückkanal");
            html << table_header("Hinkanal");
            html << table_header("Resultat");
            html << table_header("Uhrzeit");
            html << table_row_end();
        }

        while (offset <= c)
        {
            boost::shared_ptr<
                std::list<boost::shared_ptr<
                    const Database::transmission_quality_device_record> > >
                        records = db->getTransmissionQualityResults(
                            task->getTaskID(), MAX_FETCH_RECORDS, offset);
            if (records->empty())
            {
                break;
            }

            while (!records->empty())
            {
                boost::shared_ptr<
                    const Database::transmission_quality_device_record> rec =
                        records->front();
                records->pop_front();
                html << renderTQDevREcord(rec);
            }
            mg_write(m_connection, html.str().c_str(), html.str().length());
            html.clear();
            html.str("");

            offset = offset + MAX_FETCH_RECORDS;
        }

        if (c > 0)
        {
            html << table_end();
            mg_write(m_connection, html.str().c_str(), html.str().length());
        }

    }
    catch (Exception ex)
    {
        if (c > 0)
        {
            s = table_end();
            mg_write(m_connection, s.c_str(), s.length());
        }

        s = error("Could not complete report: " + ex.getMessage());
        mg_write(m_connection, s.c_str(), s.length());
    }

    s = page_end();
    mg_write(m_connection, s.c_str(), s.length());
}

void HTMLReportWebHandler::taskReport()
{
    std::string task_id = m_parameters["id"];
    if (task_id.empty())
    {
        failure("Error", "Can not generate report: missing task id parameter!");
        return;
    }

    bool dl = false;
    std::string download = m_parameters["download"];
    if (!download.empty())
    {
        try
        {
            dl = (util::str2int(download));
        }
        catch (Exception ex)
        {
            failure("Error",
                    "Can not generate report: invalid download paramter!");
            return;
        }
    }
    boost::shared_ptr<dSLongTask> task;
    try
    {
        task = Database::getInstance()->getTask(util::str2llint(task_id));
    }
    catch (Exception ex)
    {
        failure("Error", "Can not generate report! " + ex.getMessage());
        return;
    }

    switch (task->getTaskType())
    {
        case TestTransmissionQuality:
            transmissionQualityReport(task, dl);
            break;
        default:
            failure("Error", "Can not generate report: unknown task type!");
            return;
    }
}

void HTMLReportWebHandler::process()
{
    // unlike the json calls the HTML report can probably get quite big,
    // therefore we will try to keep a lesser amount of data in memory and
    // avoid passing large strings around, instead each report has to
    // write chunks of data using mg_write()
    try
    {
        if (m_command == "taskReport")
        {
            taskReport();
        }
        else
        {
            failure("Error - can not generate report!",
                    "Unknown function called");
        }
    }
    // we can't return anything to the user at this point because data is
    // written out in chunks, meaning that the user received at least a part of
    // the page already; the functions that take care of processing each call
    // have to ensure that the html never gets broken and thus handle all
    // possible exceptions gracefully
    catch (Exception ex)
    {
        fprintf(stderr, "Error generating report: %s\n",
                ex.getMessage().c_str());
    }
    catch (...)
    {
        fprintf(stderr, "Encountered unexpected error while processing call");
    }
}

} // namespace

