/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Michael Tross <mt@aizo.ag>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifdef HAVE_CONFIG_H
    #include "autoconfig.h"
#endif

#include <boost/format.hpp>
#include "exceptions.h"
#include "hwstats.h"
#include "storage.h"
#include "util.h"

namespace dSA
{


named_value_t device_sdesc_v1[] = {
{ "Version", "dataset version", 1, unsignedint },
{ "Length", "length of dataset in bytes", 1, unsignedint },
{ "Temperature", "chip temperature", 1, dsc11temp },
{ "Polarity", "polarity of receiver", 1, unsignedint },
{ "Scene", "current scene value", 1, unsignedint },
{ "LastScene", "previous scene value", 1, unsignedint },
{ "OverTempCnt", "overtemperature event counter", 2, unsignedint },
{ "OverCurrCnt", "overcurrent (shortcirc) event counter", 2, unsignedint },
{ "OverLoadCnt", "overload event counter", 2, unsignedint },
{ "NetOverLoadCnt", "net overload event counter", 2, unsignedint },
{ "RcMsgLongCnt", "count of transmitted long rev chan messages", 2, unsignedint },
{ "RcMsgShortCnt", "Count of transmitted short rev chan messages", 2, unsignedint },
{ "FcHalfWaves", "halfwaves", 4, zeroxcount },
{ "FcStartLongCnt", "start-words long", 2, unsignedint },
{ "FcStartShortCnt2", "start-words short", 2, unsignedint },
{ "FcMsgLongCnt", "correct long", 2, unsignedint },
{ "FcMsgShortCnt", "correct short", 2, unsignedint },
{ "FcErr10Cnt", "corrected bits/hamming codewords (corrected from 1 to 0)", 2, unsignedint },
{ "FcErr01Cnt", "corrected bits/hamming codewords (corrected from 0 to 1)", 2, unsignedint },
{ "FcErr2Cnt", "discarded hamming codewords", 2, unsignedint },
{ "FcBitLen[6]", "average bit length from within correct messages only excl. start word (01Beg, 01End, 10Beg, 10End, 11Beg, 11End)", 6, unsignedint },
{ "FcBitVar[6]", "variance of bit length from within correct messages only excl. start word (01Beg, 01End, 10Beg, 10End, 11Beg, 11End)", 12, unsignedint },
{ "FcRecToBeg", "#of receiver timeouts (beginning of hw)", 2, unsignedint },
{ "FcRecToEnd", "#of receiver timeouts (end of hw)", 2, unsignedint },
{ "FcRecErrBeg", "#of receiver errors (beginning of hw)", 2, unsignedint },
{ "FcRecErrEnd", "#of receiver errors (end of hw)", 2, unsignedint },
{ "FcBitNoMsgBeg", "#of received pulses (non-zeros) not within correctly decoded telegrams (beginning of hw)", 2, unsignedint },
{ "FcBitNoMsgEnd", "#of received pulses (non-zeros) not within correctly decoded telegrams (end of hw)", 2, unsignedint },
{ "FcBitMsgBeg", "#of received pulses (non-zeros) from within correctly decoded telegrams (beginning of hw)", 2, unsignedint },
{ "FcBitMsgEnd", "#of received pulses (non-zeros) from within correctly decoded telegrams (end of hw)", 2, unsignedint },
{ NULL, NULL, 0, invalid }
};

named_value_t device_sdesc_v4[] = {
{ "Version", "dataset version", 1, unsignedint },
{ "Length", "length of dataset in bytes", 1, unsignedint },
{ "Temperature", "chip temperature", 1, dsc11temp },
{ "Polarity", "polarity of receiver", 1, unsignedint },
{ "Scene", "current scene value", 1, unsignedint },
{ "LastScene", "previous scene value", 1, unsignedint },
{ "OverTempCnt", "overtemperature event counter", 2, unsignedint },
{ "OverCurrCnt", "overcurrent (shortcirc) event counter", 2, unsignedint },
{ "OverLoadCnt", "overload event counter", 2, unsignedint },
{ "NetOverLoadCnt", "net overload event counter", 2, unsignedint },
{ "RcMsgLongCnt", "count of transmitted long upstream messages", 2, unsignedint },
{ "RcMsgShortCnt", "count of transmitted short upstream messages", 2, unsignedint },
{ "RcMsgShortResendCnt", "count of repeated short upstream messages", 2, unsignedint },
{ "FcHalfWaves", "halfwaves", 4, zeroxcount },
{ "FcStartCnt", "start-words", 2, unsignedint },
{ "FcMsgLongCnt", "correct long downstream messages", 2, unsignedint },
{ "FcMsgShortCnt", "correct short downstream messages", 2, unsignedint },
{ "FcErr10Cnt", "corrected bits/hamming codewords (corrected from 1 to 0)", 2, unsignedint },
{ "FcErr01Cnt", "corrected bits/hamming codewords (corrected from 0 to 1)", 2, unsignedint },
{ "FcErr2Cnt", "discarded hamming codewords", 2, unsignedint },
{ "FcBitLenAv1", "average bit length (1st bit)", 1, unsignedint },
{ "FcBitLenAv2", "average bit length (2nd bit)", 1, unsignedint },
{ "FcBitLenMin1", "#minimal bit length (1st bit)", 1, unsignedint },
{ "FcBitLenMin2", "#minimal bit length (2nd bit)", 1, unsignedint },
{ "FcBitLenMax1", "#maximal bit length (1st bit)", 1, unsignedint },
{ "FcBitLenMax2", "#maximal bit length (2nd bit)", 1, unsignedint },
{ "FcBitIvCnt[16]", "#of bits by interval (8x 1st/2nd) (0..31/32..39/40..47/48..55/56..63/64..71/72..79/80..127)", 32, unsignedint },
{ "FcRecTo1", "#of receiver timeouts (1st bit)", 2, unsignedint },
{ "FcRecTo2", "#of receiver timeouts (2nd bit)", 2, unsignedint },
{ "FcRecErr1", "#of receiver errors (1st bit)", 2, unsignedint },
{ "FcRecErr2", "#of receiver errors (2nd bit)", 2, unsignedint },
{ "FcBitNoMsg1", "#of received pulses (non-zeros) not within correctly decoded telegrams (1st bit)", 2, unsignedint },
{ "FcBitNoMsg2", "#of received pulses (non-zeros) not within correctly decoded telegrams (2nd bit)", 2, unsignedint },
{ "FcBitMsg1", "#of received pulses (non-zeros) from within correctly decoded telegrams (1st bit)", 2, unsignedint },
{ "FcBitMsg2", "#of received pulses (non-zeros) from within correctly decoded telegrams (2nd bit)", 2, unsignedint },
{ "ChipResetCnt", "dSC11 reset event counter", 2, unsignedint },
{ "OutChangedCnt", "Count of output changes (e.g. Scene Call, local push button)", 2, unsignedint },
{ "FcHalfWavesOn", "#halfwaves device was turned on", 4, zeroxcount },
{ "LTGlitchCnt", "#of LT-Button glitch events", 1, unsignedint },
{ "LTButtonCnt", "#of LT-Button press event", 1, unsignedint },
{ "EXTGlitchCnt", "#of EXT-Button glitch events", 1, unsignedint },
{ "EXTButtonCnt", "#of EXT-Button press events", 1, unsignedint },
{ "FcDiscLongCnt", "#discarded long downstream messages", 2, unsignedint },
{ "FcDiscShortCnt", "#discarded short downstream messages", 2, unsignedint },
{ NULL, NULL, 0, invalid }
};

named_value_t meter_sdesc_v7[] = {
{ "Version", "Statistic data version", 2, unsignedint },
{ "RxCorrectShort", "Number of received short telegrams that are seemingly correct", 2, unsignedint },
{ "RxCorrectLong", "Number of received long telegrams that are seemingly correct", 2, unsignedint },
{ "RxCorrectPosHw", "Total number of correctly received telegrams in the positive mains half-wave", 2, unsignedint },
{ "RxIncorrectPosHw", "Total number of decoding failures in the positive mains half-wave", 2, unsignedint },
{ "RxErrorCrcShort", "Number of CRC errors in short frames", 2, unsignedint },
{ "RxErrorCrcLong", "Number of CRC errors in long frames", 2, unsignedint },
{ "RxErrorHeaderShort", "Number of short frames with incorrect frame ID bit", 2, unsignedint },
{ "RxErrorHeaderLong", "Number of long frames with incorrect frame ID", 2, unsignedint },
{ "AvRxQuality", "Average viterbi metric over correct frames", 2, unsignedint },
{ "AvSilencePowerOne", "Average silence power on the upper carrier", 4, unsignedint },
{ "AvSilencePowerZero", "Average silence power on the lower carrier", 4, unsignedint },
{ "AvOneBitPower", "Average signal power on the upper carrier", 4, unsignedint },
{ "AvZeroBitPower", "Average signal power on the lower carrier", 4, unsignedint },
{ "CarrierDetectPowerOnly", "Number of times the received power exceeded the carrier criterion", 4, unsignedint },
{ "CarrierDetect", "Number of times a carrier was detected", 4, unsignedint },
{ "CarrierRejectUnequal", "Number of times the carrier was rejected because the minimum bit power was too low compared to the median bit power", 4, unsignedint },
{ "HalfwaveCount", "Number of half-waves since last reboot or reset of statistics", 4, zeroxcount },
{ "SyncPosition", "Distribution of the synchronization positions", 2 * 2 * 5, unsignedint },
{ "AvZeroBitQuality", "Averaged quality of the bits received on the lower carrier", 2 * 9, unsignedint },
{ "AvOneBitQuality", "Averaged quality of the bits received on the upper carrier", 2 * 9, unsignedint },
{ "AvTotalZeroBitQualityPosHw", "Quality of the bits received on the lower carrier averaged over all bits, positive halfwave", 2, unsignedint },
{ "AvTotalZeroBitQualityNegHw", "Quality of the bits received on the lower carrier averaged over all bits, negative halfwave", 2, unsignedint },
{ "AvTotalOneBitQualityPosHw", "Quality of the bits received on the upper carrier averaged over all bits, positive halfwave", 2, unsignedint },
{ "AvTotalOneBitQualityNegHw", "Quality of the bits received on the upper carrier averaged over all bits, negative halfwave", 2, unsignedint },
{ "AvBlanked", "Average number of bit positions that have been 'blanked' by the outlier detection", 2, unsignedint },
{ "NrBlanked", "Number of bit positions that have been 'blanked'", 2, unsignedint },
{ "ExcessDecodes", "Number of decoding attempts that were made when a carrier was detected for durations that exceed the length of long telegrams", 2, unsignedint },
{ "EqualizeOneFactor", "Current equalizer gain for upper carrier", 1, unsignedint },
{ "EqualizeZeroFactor", "Current equalizer gain for upper carrier", 1, unsignedint },
{ "AvHwMismatch", "Average difference of the duration of positive and negative half-waves in units of 4/3 us", 1, signedint },
{ NULL, NULL, 0, invalid }
};


StatisticBase::StatisticBase(uint16_t version)
{
    m_version = version;
    m_index = 0;
    setDescriptionTable(NULL, 0);
    memset(m_data, 0, sizeof(m_data));
    memset(m_valid, 0, sizeof(m_valid));
}

StatisticBase::~StatisticBase() {}

uint16_t StatisticBase::getVersion() const
{
    return m_version;
}

uint8_t StatisticBase::getValue8(uint8_t index) const
{
    return m_data[index];
}

uint16_t StatisticBase::getValue16(uint8_t index) const
{
    uint16_t rval =
            (m_data[index+1] << 8) |
            (m_data[index]);
    return rval;
}

uint32_t StatisticBase::getValue32(uint8_t index) const
{
    uint32_t rval =
            (m_data[index+3] << 24) |
            (m_data[index+2] << 16) |
            (m_data[index+1] << 8) |
            (m_data[index+0]);
    return rval;
}

std::string StatisticBase::getValueString(const named_value_t *item, int offset)
{
    std::string result;

    switch (item->type)
    {
    case unsignedint:
    case signedint:
    default:
        switch (item->size)
        {
        case 1:
        {
            uint8_t val = getValue8(offset);
            result = util::hex(val, 2);
        }
        break;
        case 2:
        {
            uint16_t val = getValue16(offset);
            result = util::hex(val, 4);
        }
        break;
        case 4:
        {
            uint32_t val = getValue32(offset);
            result = util::hex(val, 8);
        }
        break;
        default:
        {
            result = util::hex(m_data + offset, item->size);
        }
        break;
        }

        break;
    case zeroxcount:
    {
        uint32_t val = getValue32(offset);
        result = util::hex(val, 8);
    }
    break;

    case dsc11temp:
    {
        uint8_t val = getValue8(offset);
        result = util::hex(val, 2);
    }
    break;
    }

    return result;
}

long StatisticBase::getValueInt(const named_value_t *item, int offset)
{
    long result;

    switch (item->type)
    {
    case unsignedint:
    case signedint:
    default:
        switch (item->size)
        {
        case 1:
            result = getValue8(offset);
            break;
        case 2:
            result = getValue16(offset);
            break;
        case 4:
            result = getValue32(offset);
            break;
        default:
            result = -1;
            break;
        }
        break;

    case zeroxcount:
        result = getValue32(offset);
        break;

    case dsc11temp:
        result = getValue8(offset);
        break;
    }

    return result;
}

void StatisticBase::pushValue(uint8_t val)
{
    m_valid[m_index] = true;
    m_data[m_index++] = val;
}

void StatisticBase::pushValue(uint16_t val)
{
    pushValue( (uint8_t) (val & 0xff));
    pushValue( (uint8_t) (val >> 8));
}

void StatisticBase::pushNilValue(uint8_t num)
{
    m_index += num;
}

void StatisticBase::setDescriptionTable(const named_value_t *table, int length)
{
    m_definition_table = (named_value_t *) table;
    m_definition_table_length = length;
}

boost::shared_ptr<std::list<named_value_t> > StatisticBase::getParameterList()
{
    boost::shared_ptr<std::list<named_value_t> > result(
                                                 new std::list<named_value_t>);
    named_value_t* t;

    for (t = m_definition_table; t->name; t ++)
    {
        result->push_back(*t);
    }
    return result;
}

boost::shared_ptr<std::list<std::string> > StatisticBase::getValueList()
{
    boost::shared_ptr<std::list<std::string> > result(
                                                   new std::list<std::string>);
    named_value_t* t;
    int offset;
    bool incomplete = false;

    for (t = m_definition_table, offset = 0; t->name; offset += t->size, t++)
    {
        for (int n = 0; n < t->size; n ++)
        {
            if (!m_valid[offset+n])
            {
                incomplete = true;
            }
        }
        if (incomplete)
        {
            result->push_back(std::string("incomplete"));
            continue;
        }

        result->push_back(getValueString(t, offset));
    }
    return result;
}

boost::shared_ptr<std::list<std::string> > StatisticBase::genResultList()
{
    boost::shared_ptr<std::list<std::string> > result(
                                                   new std::list<std::string>);
    named_value_t *t;
    int offset;
    bool incomplete = false;

    for (t = m_definition_table, offset = 0;
            t->name;
            offset += t->size, t++)
    {
        for (int n = 0; n < t->size; n ++)
        {
            if (!m_valid[offset+n])
            {
                incomplete = true;
            }
        }
        if (incomplete)
        {
            result->push_back(std::string(t->name) + "=incomplete");
            continue;
        }

        switch (t->type)
        {
            case unsignedint:
            case signedint:
            default:
                result->push_back(std::string(t->name) + "=" +
                        getValueString(t, offset));
                break;

            case zeroxcount:
                {
                    uint32_t val = getValue32(offset);
                    uint32_t seconds = (val + 50) / 100;
                    uint32_t minutes = seconds / 60; seconds %= 60;
                    uint32_t hours = minutes / 60; minutes %= 60;
                    uint32_t days = hours / 24; hours %= 24;
                    std::string uptime(boost::str(
                        boost::format(
                            "%1%d %2%h %3%m %4%s"
                            ) % days % hours % minutes % seconds ));
                    result->push_back(std::string(t->name) + "=" +
                                util::hex(val, 8) + " (" + uptime + ")" );
                }
                break;

            case dsc11temp:
                {
                    uint8_t val = getValue8(offset);
                    std::string celsius(boost::str(
                        boost::format("%1% °C, +/-10") % (149 - 0.9 * val)));
                    result->push_back(std::string(t->name) + "=" +
                            util::hex(val, 2) + " (" + celsius + ")");
                }
                break;
        }
    }
    return result;
}


StatisticDevice::StatisticDevice(uint8_t version) : StatisticBase(version)
{
    // select defintion table based on version
    switch (version) {
        case 0x01:
        case 0x02:
            setDescriptionTable(device_sdesc_v1,
                    sizeof(device_sdesc_v1) / sizeof(named_value_t) - 1);
            break;
        case 0x03:
        case 0x04:
        default:
            setDescriptionTable(device_sdesc_v4,
                    sizeof(device_sdesc_v4) / sizeof(named_value_t) - 1);
            break;
    }
}

StatisticMeter::StatisticMeter(uint16_t version) : StatisticBase(version)
{
    // select defintion table based on version
    switch (version)
    {
        case 0x07:
        default:
            setDescriptionTable(meter_sdesc_v7,
                    sizeof(meter_sdesc_v7) / sizeof(named_value_t) - 1);
            break;
    }
}

}
