/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Michael Tross <mt@aizo.ag>, based on http://www.drdobbs.com/cpp/201804215

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifndef __LOG_H__
#define __LOG_H__

#include <sstream>
#include <string>
#include <stdio.h>
#include <sys/time.h>
#include <boost/thread/thread.hpp>
#include <boost/thread/mutex.hpp>
#include "util.h"

namespace dSA
{

inline std::string NowTime()
{
    char buffer[32];
    time_t t;
    time(&t);
    struct tm r = { 0 };
    strftime(buffer, sizeof(buffer), "%x %X", localtime_r(&t, &r));
    struct timeval tv;
    gettimeofday(&tv, 0);
    char result[100];
    sprintf(result, "%s.%03ld", buffer, (long) tv.tv_usec / 1000);
    return result;
}

enum TLogLevel
{
    logERROR,
    logWARNING,
    logINFO,
    logDEBUG,
    logDEBUG1,
    logDEBUG2,
    logDEBUG3,
    logDEBUG4
};


template<typename T>
class Log
{
public:
    Log();
    virtual ~Log();
    std::ostringstream& Get(TLogLevel level = logINFO);
    std::ostringstream& GetDSID(const std::string dsid,
                                TLogLevel level = logINFO);
    std::ostringstream& GetDSID(const dsid_t dsid,
                                TLogLevel level = logINFO);
    std::ostringstream& GetDSIDPrefix(const dsid_t dsid, const char *prefix,
                                TLogLevel level = logINFO);
public:
    static TLogLevel& ReportingLevel();
    static std::string ToString(TLogLevel level);
    static TLogLevel FromString(const std::string& level);
protected:
    std::ostringstream os;
private:
    Log(const Log&);
    Log& operator =(const Log&);
};

template<typename T>
Log<T>::Log()
{
}

template<typename T>
std::ostringstream& Log<T>::Get(TLogLevel level)
{
    os << "[" << NowTime() << "]";
    os << " " << ToString(level) << ": ";
    os << std::string(level > logDEBUG ? level - logDEBUG : 0, '\t');
    return os;
}

template<typename T>
std::ostringstream& Log<T>::GetDSID(const std::string dsid, TLogLevel level)
{
    os << "[" << NowTime() << "]";
    //os << "[" << dsid << "]";
    os << " " << ToString(level) << ": ";
    os << std::string(level > logDEBUG ? level - logDEBUG : 0, '\t');
    T::SetDSID(dsid);
    return os;
}

template<typename T>
std::ostringstream& Log<T>::GetDSID(const dsid_t dsid, TLogLevel level)
{
    os << "[" << NowTime() << "]";
    ///os << "[" << util::str(dsid) << "]";
    os << " " << ToString(level) << ": ";
    os << std::string(level > logDEBUG ? level - logDEBUG : 0, '\t');
    T::SetDSID(util::str(dsid));
    return os;
}

template<typename T>
std::ostringstream& Log<T>::GetDSIDPrefix(const dsid_t dsid, const char *prefix, TLogLevel level)
{
    os << "[" << NowTime() << "]";
    ///os << "[" << util::str(dsid) << "]";
    os << " " << ToString(level) << ": ";
    os << std::string(level > logDEBUG ? level - logDEBUG : 0, '\t');
    T::SetDSIDPrefix(util::str(dsid), prefix);
    return os;
}

template<typename T>
Log<T>::~Log()
{
    os << std::endl;
    T::Output(os.str());
}

template<typename T>
TLogLevel& Log<T>::ReportingLevel()
{
    static TLogLevel reportingLevel = logDEBUG4;
    return reportingLevel;
}

template<typename T>
std::string Log<T>::ToString(TLogLevel level)
{
    static const char* const buffer[] =
    {
        "ERROR", "WARNING", "INFO", "DEBUG", "DEBUG1", "DEBUG2", "DEBUG3",
        "DEBUG4"
    };
    return buffer[level];
}

template<typename T>
TLogLevel Log<T>::FromString(const std::string& level)
{
    if (level == "DEBUG4")
        return logDEBUG4;
    if (level == "DEBUG3")
        return logDEBUG3;
    if (level == "DEBUG2")
        return logDEBUG2;
    if (level == "DEBUG1")
        return logDEBUG1;
    if (level == "DEBUG")
        return logDEBUG;
    if (level == "INFO")
        return logINFO;
    if (level == "WARNING")
        return logWARNING;
    if (level == "ERROR")
        return logERROR;
    Log<T> ().Get(logWARNING) << "Unknown logging level '" << level
            << "'. Using INFO level as default.";
    return logINFO;
}

class Output2FILE
{
public:
    static void Output(const std::string& msg);
    static void SetStream(FILE* pFile);
private:
    static FILE*& StreamImpl();
    static boost::mutex& MutexImpl();
};

inline FILE*& Output2FILE::StreamImpl()
{
    static FILE* pStream = stderr;
    return pStream;
}

inline boost::mutex& Output2FILE::MutexImpl()
{
    static boost::mutex m_mutex;
    return m_mutex;
}

inline void Output2FILE::SetStream(FILE* pFile)
{
    boost::mutex::scoped_lock lock(MutexImpl());
    StreamImpl() = pFile;
}

inline void Output2FILE::Output(const std::string& msg)
{
    boost::mutex::scoped_lock lock(MutexImpl());
    FILE* pStream = StreamImpl();
    if (!pStream)
        return;
    fprintf(pStream, "%s", msg.c_str());
    fflush(pStream);
}

class FILELog: public Log<Output2FILE>
{
};


class Output2DSID
{
public:
    static void Output(const std::string& msg);
    static void SetDSID(const std::string dsid);
    static void SetDSIDPrefix(const std::string dsid, const char *prefix);
    static void SetLogdirectory(const std::string dir);
private:
    static FILE*& StreamImpl();
    static boost::mutex& MutexImpl();
    static std::string& LogdirImpl();
};

inline FILE*& Output2DSID::StreamImpl()
{
    static FILE* pStream = stderr;
    return pStream;
}

inline boost::mutex& Output2DSID::MutexImpl()
{
    static boost::mutex m_mutex;
    return m_mutex;
}

inline std::string& Output2DSID::LogdirImpl()
{
    static std::string m_logdir;
    return m_logdir;
}

inline void Output2DSID::SetDSID(const std::string dsid)
{
    boost::mutex::scoped_lock lock(MutexImpl());
    std::string filename(LogdirImpl() + "/dsid_" +
                         dsid.c_str() + std::string(".log"));
    StreamImpl() = fopen(filename.c_str(), "a");
}

inline void Output2DSID::SetDSIDPrefix(const std::string dsid, const char *prefix)
{
    boost::mutex::scoped_lock lock(MutexImpl());
    std::string filename(LogdirImpl() + "/" + std::string(prefix) +
                         dsid.c_str() + std::string(".log"));
    StreamImpl() = fopen(filename.c_str(), "a");
}

inline void Output2DSID::SetLogdirectory(const std::string dir)
{
    LogdirImpl() = dir;
}

inline void Output2DSID::Output(const std::string& msg)
{
    boost::mutex::scoped_lock lock(MutexImpl());
    FILE* pStream = StreamImpl();
    if (!pStream)
        return;
    fprintf(pStream, "%s", msg.c_str());
    fclose(pStream);
}

class DSIDLog: public Log<Output2DSID>
{
};

#ifndef FILELOG_MAX_LEVEL
#ifdef NDEBUG
# define FILELOG_MAX_LEVEL logDEBUG
#else
# define FILELOG_MAX_LEVEL logDEBUG4
#endif
#endif

#define FILE_LOG(level) \
    if (level > FILELOG_MAX_LEVEL) ;\
    else if (level > FILELog::ReportingLevel()) ; \
    else FILELog().Get(level)

#ifndef DSID_MAX_LEVEL
#ifdef NDEBUG
# define DSID_MAX_LEVEL logDEBUG
#else
# define DSID_MAX_LEVEL logDEBUG4
#endif
#endif

#define DSID_LOG(level, dsid) \
    if (level > DSID_MAX_LEVEL) ;\
    else if (level > DSIDLog::ReportingLevel()) ; \
    else DSIDLog().GetDSID(dsid, level)

#define DSID_PREFIX_LOG(_level, _dsid, _prefix) \
    if (_level > DSID_MAX_LEVEL) ;\
    else if (_level > DSIDLog::ReportingLevel()) ; \
    else DSIDLog().GetDSIDPrefix(_dsid, _prefix, _level)

};

#endif //__LOG_H__
